"""
Tests for plot and animation helper functions
"""

from unittest import TestCase

import xarray as xr
import numpy as np
import pandas as pd

from esa_climate_toolbox.ops.plot_helpers import check_bounding_box
from esa_climate_toolbox.ops.plot_helpers import determine_cmap_params
from esa_climate_toolbox.ops.plot_helpers import get_var_data
from esa_climate_toolbox.ops.plot_helpers import get_vars_data
from esa_climate_toolbox.ops.plot_helpers import in_notebook
from esa_climate_toolbox.core.types import ValidationError


class TestCheckBoundingBox(TestCase):
    """
    Test check_bounding_box()
    """

    def test_nominal(self):
        """
        Nominal test
        """
        self.assertTrue(check_bounding_box(10, 20, 10, 20))
        self.assertFalse(check_bounding_box(20, 10, 10, 20))
        self.assertFalse(check_bounding_box(10, 20, 20, 10))
        self.assertFalse(check_bounding_box(-91, 10, 10, 20))
        self.assertFalse(check_bounding_box(20, 91, 10, 20))
        self.assertFalse(check_bounding_box(10, 20, -181, 20))
        self.assertFalse(check_bounding_box(10, 20, 10, 181))


class TestInNotebook(TestCase):
    """
    Test in_notebook()
    """

    def test_nominal(self):
        """
        Nominal test
        """
        self.assertFalse(in_notebook())


class TestGetVarData(TestCase):
    """
    Test get_var_data()
    """

    def test_nominal(self):
        """
        Nominal test
        """
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'second': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'lat': np.linspace(-89.5, 89.5, 5),
            'lon': np.linspace(-179.5, 179.5, 10),
            'time': pd.date_range('2000-01-01', periods=2)})

        indexers = {'time': '2000-01-01', 'lat': -88}
        out = get_var_data(dataset.second, indexers, remaining_dims=['lon', ])
        self.assertEqual(out.time, np.datetime64('2000-01-01'))
        self.assertEqual(out.lat, -89.5)
        self.assertEqual(len(out.lon.values), 10)

        # test on data with more dims
        dataset2 = xr.Dataset({
            'first': (['lat', 'lon', 'layers', 'time'], np.random.rand(5, 10, 16, 2)),
            'second': (['lat', 'lon', 'layers', 'time'], np.random.rand(5, 10, 16, 2)),
            'lat': np.linspace(-89.5, 89.5, 5),
            'lon': np.linspace(-179.5, 179.5, 10),
            'layers': np.linspace(1, 16, 16),
            'time': pd.date_range('2000-01-01', periods=2)})

        indexers = {'lat': 0, 'lon': -179.5, 'layers': 5}
        out = get_var_data(dataset2.second, indexers, remaining_dims=['time', ])
        self.assertEqual(out.lat, 0)
        self.assertEqual(out.lon, -179.5)
        self.assertEqual(out.layers, 5)
        self.assertEqual(len(out.time.values), 2)

        # test when a dimension is neither in indexers nor remaining_dims specified,
        # it will select the first element of that dimension
        # (if remaining_dims is specified)
        indexers = {'lat': 0, 'lon': -179.5}
        out = get_var_data(dataset2.second, indexers, remaining_dims=['time', ])
        self.assertEqual(out.lat, 0)
        self.assertEqual(out.lon, -179.5)
        self.assertEqual(out.layers, 1)
        self.assertEqual(len(out.time.values), 2)

        # when remaining_dims is not specified,
        # it should only index based on the given indexers
        indexers = {'lat': 0, 'lon': -179.5}
        out = get_var_data(dataset2.second, indexers)
        self.assertEqual(out.lat, 0)
        self.assertEqual(out.lon, -179.5)
        self.assertEqual(len(out.layers), 16)
        self.assertEqual(len(out.time.values), 2)

        # should raise a ValidationError when
        # the specified dimension is not part of variable dimension names
        indexers = {'lat': 0, 'lon': -179.5, 'time': '2000-01-01'}
        with self.assertRaises(ValidationError) as cm:
            get_var_data(dataset2.second, indexers, remaining_dims=['dummy', ])
        self.assertEqual(str(cm.exception),
                         "The specified dataset does not have a "
                         "dimension called \'dummy\'.")

        # should raise a ValidationError when a dimension is specified
        # as both indexers and remaining_dims
        indexers = {'lat': 0, 'lon': -179.5, 'time': '2000-01-01'}
        with self.assertRaises(ValidationError) as cm:
            get_var_data(dataset2.second, indexers, remaining_dims=['time', ])
        self.assertEqual(str(cm.exception),
                         "Dimension 'time' is also specified as indexers. "
                         "Please ensure that a dimension is used exclusively "
                         "either as indexers or as the selected dimension.")

        # should raise a ValidationError when a dimension is specified
        # as indexers does not exist in the given variable
        indexers = {'lat': 0, 'lon': -179.5, 'dummy': 1}
        with self.assertRaises(ValidationError) as cm:
            get_var_data(dataset2.second, indexers, remaining_dims=['time', ])
        self.assertEqual(str(cm.exception),
                         "The specified dataset does not have a dimension "
                         "called 'dummy'.")

    def test_get_vars_data(self):
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'layers1', 'time'], np.random.rand(5, 10, 16, 2)),
            'second': (['lat', 'lon', 'layers2', 'time'], np.random.rand(5, 10, 4, 2)),
            'third': (['lat', 'lon', 'layers1', 'layers2', 'time'],
                      np.random.rand(5, 10, 16, 4, 2)),
            'fourth': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'lat': np.linspace(-89.5, 89.5, 5),
            'lon': np.linspace(-179.5, 179.5, 10),
            'layers1': np.linspace(1, 16, 16),
            'layers2': np.linspace(1, 4, 4),
            'time': pd.date_range('2000-01-01', periods=2)})

        indexers = {'lat': -89.5, 'lon': -179.5, 'layers1': 5, 'layers2': 4}
        vars_data = get_vars_data(dataset, indexers, remaining_dims=['time', ])
        self.assertEqual(9, len(vars_data.variables))
        self.assertEqual(5, len(vars_data.dims))
        self.assertEqual(1, len(vars_data.first.dims))
        self.assertEqual(1, len(vars_data.second.dims))
        self.assertEqual(1, len(vars_data.third.dims))
        self.assertEqual(1, len(vars_data.fourth.dims))
        # TODO (hp): is it possible to check if the right data has been selected?
        #  When merged back to ds, the dimensions selected by the indexers
        #  are not visible anymore.
        self.assertEqual(2, len(vars_data.time))

        indexers = {'lat': -89.5, 'lon': 179.5}
        vars_data = get_vars_data(dataset, indexers, remaining_dims=['time', ])
        self.assertEqual(9, len(vars_data.variables))
        self.assertEqual(5, len(vars_data.dims))
        self.assertEqual(1, len(vars_data.first.dims))
        self.assertEqual(1, len(vars_data.second.dims))
        self.assertEqual(1, len(vars_data.third.dims))
        self.assertEqual(1, len(vars_data.fourth.dims))
        self.assertEqual(2, len(vars_data.time))

        indexers = {'lat': -89.5, 'lon': 179.5}
        vars_data = get_vars_data(dataset, indexers)
        self.assertEqual(9, len(vars_data.variables))
        self.assertEqual(5, len(vars_data.dims))
        self.assertEqual(2, len(vars_data.first.dims))
        self.assertEqual(2, len(vars_data.second.dims))
        self.assertEqual(3, len(vars_data.third.dims))
        self.assertEqual(1, len(vars_data.fourth.dims))
        self.assertEqual(2, len(vars_data.time))

        # should raise a ValidationError when the specified dimension
        # is not part of variable dimension names
        indexers = {'lat': 0, 'lon': -179.5, 'time': '2000-01-01'}
        with self.assertRaises(ValidationError) as cm:
            get_vars_data(dataset, indexers, remaining_dims=['dummy', ])
        self.assertEqual("The specified dataset does not have a dimension "
                         "called 'dummy'.", str(cm.exception))

        # should raise a ValidationError when a dimension is specified
        # as both indexers and remaining_dims
        indexers = {'lat': 0, 'lon': -179.5, 'time': '2000-01-01'}
        with self.assertRaises(ValidationError) as cm:
            get_vars_data(dataset, indexers, remaining_dims=['time', ])
        self.assertEqual("Dimension 'time' is also specified as indexers. "
                         "Please ensure that a dimension is used exclusively either "
                         "as indexers or as the selected dimension.", str(cm.exception))

        # should raise a ValidationError when a dimension is specified
        # as indexers does not exist in the given variable
        indexers = {'lat': 0, 'lon': -179.5, 'dummy': 1}
        with self.assertRaises(ValidationError) as cm:
            get_vars_data(dataset, indexers, remaining_dims=['time', ])
        self.assertEqual("There are dimensions specified in indexers "
                         "but do not match dimensions in any variables: ['dummy']",
                         str(cm.exception))


class TestDetermineCmapParams(TestCase):
    """
    Test determine_cmap_params()
    """

    def test_nominal(self):
        """
        Nominal test
        """
        # Test nominal
        cparams = determine_cmap_params(0, 100)
        self.assertEqual(cparams['cmap'], 'viridis')

        # Test divergent
        cparams = determine_cmap_params(-100, 100)
        self.assertEqual(cparams['cmap'], 'RdBu_r')

        # Test extents
        cparams = determine_cmap_params(0, 100, vmin=10, vmax=90)
        self.assertEqual(cparams['extend'], 'both')

        # Test levels
        cparams = determine_cmap_params(0, 100, levels=3, vmin=10)
        self.assertEqual(len(cparams['levels']), 3)
        self.assertIsNotNone(cparams['norm'])



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import datetime
import geopandas as gpd
from typing import Any
from typing import Container
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
import xarray as xr
import yaml

from xcube.core.mldataset import MultiLevelDataset
from xcube.core.select import select_subset
import xcube.core.store as xcube_store
from xcube.util.assertions import assert_instance
from xcube.util.progress import add_progress_observers
from xcube.util.progress import ProgressObserver
from xcube.util.progress import ProgressState

from esa_climate_toolbox.conf.defaults import STORES_CONF_FILE
from ..constants import ECT_STORE_ID
from ..constants import ECT_ZARR_STORE_ID
from .types import PolygonLike
from .types import TimeRangeLike
from .types import ValidationError
from .types import VarNamesLike
from ..util.monitor import ChildMonitor
from ..util.monitor import Monitor

_DEFAULT_FILE_STORE_ID_PREFIX = 'local'
_OUTPUT_STORE = None

ECT_DATA_STORE_POOL = xcube_store.DataStorePool()


class DataAccessError(Exception):
    """
    Exceptions produced by the data stores of the ESA Climate Toolbox,
    used to report any problems handling data.
    """


class XcubeProgressObserver(ProgressObserver):

    def __init__(self, monitor: Monitor):
        self._monitor = monitor
        self._latest_completed_work = 0.0

    def on_begin(self, state_stack: Sequence[ProgressState]):
        if len(state_stack) == 1:
            self._monitor.start(state_stack[0].label, state_stack[0].total_work)

    def on_update(self, state_stack: Sequence[ProgressState]):
        if state_stack[0].completed_work > self._latest_completed_work:
            self._monitor.progress(state_stack[0].completed_work
                                   - self._latest_completed_work,
                                   state_stack[-1].label)
            self._latest_completed_work = state_stack[0].completed_work

    def on_end(self, state_stack: Sequence[ProgressState]):
        if len(state_stack) == 1:
            self._monitor.done()


def list_stores() -> List[str]:
    """
    Lists the names of the data stores which are provided by
    Returns meta information about an operation.

    :param op_name: The name of the operation for which meta information shall
        be provided.
    :param op_registry: An optional OpRegistry, in case the default one should
        not be used.

    :return: A dictionary representation of an operator's meta info,
        providing information about input parameters and the expected output.
    """
    return ECT_DATA_STORE_POOL.store_instance_ids


def list_ecvs() -> List[str]:
    """
    Returns a list of names of essential climate variables served by the
    ESA Climate Toolbox.

    :return: A list of names of essential climate variables.
    """
    store = ECT_DATA_STORE_POOL.get_store(ECT_STORE_ID)
    data_ids = store.list_data_ids()
    return list(set([data_id.split('.')[1] for data_id in data_ids]))


def list_ecv_datasets(
        ecv: str,
        data_type: xcube_store.DataTypeLike = None,
        include_attrs: Container[str] = None
) -> Union[List[str], List[Tuple[str, Dict[str, Any]]]]:
    """
    Returns the names of datasets for a given essential climate variable.

    :param ecv: The name of the essential climate variable
    :param data_type: A datatype that may be provided to restrict the search,
        e.g., 'dataset' or 'geodataframe'
    :param include_attrs: An optional list to retrieve
        additional meta information if required.

    :return: Either a list of dataset names for the given ecv, or a list of
        tuples, each consisting of a name and a dictionary with additional
        information.
    """
    ecvs = list_ecvs()
    if ecv.upper() not in ecvs:
        raise ValueError(f'"{ecv}" is not an Essential Climate Variable '
                         f'provided by the ESA Climate Toolbox. Please choose '
                         f'one of the following: {", ".join(ecvs)}')
    ecv = ecv.upper()
    ecv_datasets = []
    for store_instance_id in ECT_DATA_STORE_POOL.store_instance_ids:
        store = ECT_DATA_STORE_POOL.get_store(store_instance_id)
        data_ids = store.list_data_ids(data_type, include_attrs)
        for data_id in data_ids:
            if ecv in data_id.upper():
                ecv_datasets.append((data_id, store_instance_id))
    return ecv_datasets


def _get_store_id_from_prefix(prefix: str) -> str:
    count = 1
    store_id = f'{prefix}_{count}'
    while ECT_DATA_STORE_POOL.has_store_instance(store_id):
        count += 1
        store_id = f'{prefix}_{count}'
    return store_id


def _save_store_pool():
    with open(STORES_CONF_FILE, 'w') as fp:
        yaml.safe_dump(ECT_DATA_STORE_POOL.to_dict(), fp)


def add_local_store(root: str, store_id: str = None, max_depth: int = 1,
                    read_only: bool = False, includes: str = None,
                    excludes: str = None, title: str = None,
                    description: str = None, persist: bool = True) -> str:
    """
    Registers a new data store in the ESA Climate Toolbox to access locally
    stored data.

    :param root: The path to the data.
    :param store_id: The name the store should have. There must not already
        be a store of the same name.
    :param max_depth: The maximum level of sub-directories that will be
        browsed for data. Default is 1, i.e., only the data located in the root
        path will be considered.
    :param read_only: Whether the store is read-only. Default is false.
    :param includes: Allows to specify a pattern about which data shall be
        served by the store (e.g., aerosol*.nc)
    :param excludes: Allows to specify a pattern about which data shall not be
        served by the store (e.g., aerosol*.zarr)
    :param title: An optional title for the data store
    :param description: An optional description of the data store
    :param persist: Whether the data store shall be registered permanently,
        otherwise it will only be for this session. Default is True.

    :return: The id of the newly created store.
    """
    if store_id is not None and \
            ECT_DATA_STORE_POOL.has_store_instance(store_id):
        raise ValueError(
            f'There is already a store with the id "{store_id}" registered.'
            f' Please choose another id.'
        )
    if store_id is None:
        store_id = _get_store_id_from_prefix(_DEFAULT_FILE_STORE_ID_PREFIX)
    store_params = {
        'root': root,
        'max_depth': max_depth,
        'read_only': read_only,
        'includes': includes,
        'excludes': excludes
    }
    store_config = xcube_store.DataStoreConfig(
        store_id='file', store_params=store_params,
        title=title, description=description
    )
    ECT_DATA_STORE_POOL.add_store_config(store_id, store_config)
    if persist:
        _save_store_pool()
    return store_id


def add_store(
        store_type: str, store_params: Mapping[str, Any] = None,
        store_id: str = None, title: str = None, description: str = None,
        user_data: Any = None, persist: bool = True
) -> str:
    """
    Registers a new data store in the ESA Climate Toolbox. This function allows
    to also specify non-local data stores.

    :param store_type: The type of data store to create, e.g., 's3'.
    :param store_params: A mapping containing store-specific parameters which
        are required to initiate the store.
    :param store_id: The name the store should have. There must not already
        be a store of the same name.
    :param title: An optional title for the data store
    :param description: An optional description of the data store
    :param user_data: Any additional user data
    :param persist: Whether the data store shall be registered permanently,
        otherwise it will only be for this session. Default is True.

    :return: The id of the newly created store.
    """
    if store_id is not None and \
            ECT_DATA_STORE_POOL.has_store_instance(store_id):
        raise ValueError(
            f'There is already a store with the id "{store_id}" registered.'
            f' Please choose another id.'
        )
    if store_id is None:
        store_id = _get_store_id_from_prefix(store_type)
    store_config = xcube_store.DataStoreConfig(
        store_id=store_type, store_params=store_params, title=title,
        description=description, user_data=user_data
    )
    ECT_DATA_STORE_POOL.add_store_config(store_id, store_config)
    if persist:
        _save_store_pool()
    return store_id


def get_store(store_id: str):
    """
    Returns the data store of the given name.
    :param store_id: The name of the store should have.

    :return: A data store
    """
    return ECT_DATA_STORE_POOL.get_store(store_id)


def list_datasets(
        store_id: str = None,
        data_type: xcube_store.DataTypeLike = None,
        include_attrs: Container[str] = None
) -> Union[List[str], List[Tuple[str, Dict[str, Any]]]]:
    """
    Returns the names of datasets of a given store.

    :param store_id: The name of the data store
    :param data_type: A datatype that may be provided to restrict the search,
        e.g., 'dataset' or 'geodataframe'
    :param include_attrs: An optional list to retrieve
        additional meta information if required.

    :return: Either a list of the dataset names within the store, or a list of
        tuples, each consisting of a name and a dictionary with additional
        information.
    """
    if store_id:
        return ECT_DATA_STORE_POOL.get_store(store_id).list_data_ids(
            data_type, include_attrs
        )
    datasets = []
    for store_instance_id in ECT_DATA_STORE_POOL.store_instance_ids:
        store = ECT_DATA_STORE_POOL.get_store(store_instance_id)
        datasets.extend(store.list_data_ids(data_type, include_attrs))
    return datasets


def remove_store(store_id: str, persist: bool = True):
    """
    Removes a store from the internal store registry. No actual data will be
    deleted.

    :param store_id: The name of the store to be removed
    :param persist: Whether the data store shall be unregistered permanently,
        otherwise it will only be for this session. Default is True.

    :return: Either a list of dataset names for the given ecv, or a list of
        tuples, each consisting of a name and a dictionary with additional
        information.
    """
    if store_id in [ECT_STORE_ID, ECT_ZARR_STORE_ID]:
        raise ValueError(f'Cannot remove essential store "{store_id}".')
    if store_id not in ECT_DATA_STORE_POOL.store_instance_ids:
        raise ValueError(f'No store named "{store_id}" found.')
    ECT_DATA_STORE_POOL.remove_store_config(store_id)
    if persist:
        _save_store_pool()


def get_output_store_id() -> Optional[str]:
    """
    Returns the name of the store that by default will be used for writing.

    :return: The id of the default output store.
    """
    return _OUTPUT_STORE


def set_output_store(store_id: str):
    """
    Specifies which store shall be the standard output store. This value is not
    persisted and must be set every session.

    :param store_id: The name of the store that shall be the output store.
    """
    if store_id not in ECT_DATA_STORE_POOL.store_instance_ids:
        raise ValueError(f'Could not set "{store_id}" as default output store. '
                         f'Store is not registered.')
    global _OUTPUT_STORE
    _OUTPUT_STORE = store_id


def find_data_store(
        ds_id: str
) -> Tuple[Optional[str], Optional[xcube_store.DataStore]]:
    """
    Find the data store that includes the given *ds_id*.
    This will raise an exception if the *ds_id* is given
    in more than one data store.

    :param ds_id:  A data source identifier.
    :return: All data sources matching the given constrains.
    """
    results = []
    for store_instance_id in ECT_DATA_STORE_POOL.store_instance_ids:
        data_store = ECT_DATA_STORE_POOL.get_store(store_instance_id)
        if data_store.has_data(ds_id):
            results.append((store_instance_id, data_store))
    if len(results) > 1:
        raise ValidationError(
            f'{len(results)} data sources found for the given ID {ds_id!r}'
        )
    if len(results) == 1:
        return results[0]
    return None, None


def open_data(dataset_id: str,
              time_range: TimeRangeLike.TYPE = None,
              region: PolygonLike.TYPE = None,
              var_names: VarNamesLike.TYPE = None,
              data_store_id: str = None,
              monitor: Monitor = Monitor.NONE) -> Tuple[Any, str]:
    """
    Open a dataset from a data store.

    :param dataset_id: The identifier of the dataset. Must not be empty.
    :param time_range: An optional time constraint comprising start and end
        date. If given, it must be a :py:class:`TimeRangeLike`.
    :param region: An optional region constraint.
        If given, it must be a :py:class:`PolygonLike`.
    :param var_names: Optional names of variables to be included.
        If given, it must be a :py:class:`VarNamesLike`.
    :param data_store_id: Optional data store identifier. If given, *ds_id*
        will only be looked up from the specified data store.
    :param monitor: A progress monitor
    :return: A tuple consisting of a new dataset instance and its id
    """
    if not dataset_id:
        raise ValidationError('No data source given')

    if data_store_id:
        data_store = ECT_DATA_STORE_POOL.get_store(data_store_id)
    else:
        data_store_id, data_store = find_data_store(ds_id=dataset_id)
        if not data_store:
            raise ValidationError(
                f"No data store found that contains the ID '{dataset_id}'"
            )

    data_type = None
    potential_data_types = data_store.get_data_types_for_data(dataset_id)
    for potential_data_type in potential_data_types:
        if xcube_store.DATASET_TYPE.is_super_type_of(potential_data_type):
            data_type = potential_data_type
            break
    if data_type is None:
        raise ValidationError(f"Could not open '{dataset_id}' as dataset.")
    openers = data_store.get_data_opener_ids(dataset_id, data_type)
    if len(openers) == 0:
        raise DataAccessError(f'Could not find an opener for "{dataset_id}".')
    opener_id = openers[0]

    open_work = 10
    subset_work = 0

    open_schema = data_store.get_open_data_params_schema(dataset_id, opener_id)
    open_args = {}

    subset_args = {}
    if var_names:
        var_names_list = VarNamesLike.convert(var_names)
        if 'variable_names' in open_schema.properties:
            open_args['variable_names'] = var_names_list
        elif 'drop_variables' in open_schema.properties:
            data_desc = data_store.describe_data(dataset_id, data_type)
            if hasattr(data_desc, 'data_vars') \
                    and isinstance(getattr(data_desc, 'data_vars'), dict):
                open_args['drop_variables'] = [
                    var_name for var_name in data_desc.data_vars.keys()
                    if var_name not in var_names_list
                ]
        else:
            subset_args['var_names'] = var_names_list
            subset_work += 1

    if time_range:
        time_range = TimeRangeLike.convert(time_range)
        time_range = [datetime.datetime.strftime(time_range[0], '%Y-%m-%d'),
                      datetime.datetime.strftime(time_range[1], '%Y-%m-%d')]
        if 'time_range' in open_schema.properties:
            open_args['time_range'] = time_range
        else:
            subset_args['time_range'] = time_range
            subset_work += 1

    if region:
        bbox = list(PolygonLike.convert(region).bounds)
        if 'bbox' in open_schema.properties:
            open_args['bbox'] = bbox
        else:
            subset_args['bbox'] = bbox
            subset_work += 1

    with monitor.starting('Open dataset', open_work + subset_work):
        with add_progress_observers(
                XcubeProgressObserver(ChildMonitor(monitor, open_work))
        ):
            dataset = data_store.open_data(
                data_id=dataset_id, opener_id=opener_id, **open_args
            )

        dataset = select_subset(dataset, **subset_args)
        monitor.progress(subset_work)

    return dataset, dataset_id


def get_supported_formats(data: Any, store_id: str) -> List[str]:
    """
    Returns the list of formats to which the store at the given store_id may
    write the given data.

    :param data: The data which shall be written
    :param store_id: The id of the store to which the data shall be written

    :return: A list of supported output formats
    """
    if store_id is not None and \
            store_id not in ECT_DATA_STORE_POOL.store_instance_ids:
        raise ValueError(f'Unknown Data Store "{store_id}".')
    store = ECT_DATA_STORE_POOL.get_store(store_instance_id=store_id)
    assert isinstance(store, xcube_store.MutableDataStore)
    data_type = _get_data_type(data)
    writer_ids = store.get_data_writer_ids(data_type)
    return [writer_id.split(':')[1] for writer_id in writer_ids]


def write_data(
        data: Any, data_id: str = None, store_id: str = None,
        format_id: str = None, replace: bool = False,
        monitor: Monitor = Monitor.NONE
) -> str:
    """
    Writes data

    :param data: The data which shall be written
    :param data_id: A data id under which the data shall be written to the store.
        If not given, a data id will be created.
    :param store_id: The id of the store to which the data shall be written.
        If none is given, the data is written to the standard output store.
    :param format_id:
        A format that shall be used to write the data. If none is given, the
        data will be written in the default format for the data type,
        e.g., 'zarr' for datasets.
    :param replace: Whether a dataset with the same id in the store shall be
        replaced. If False, an exception will be raised. Default is False.
    :param monitor: A monitor to measure the writing process

    :return: The data id under which the data can be accessed from the store.
    """
    if store_id is not None and \
            store_id not in ECT_DATA_STORE_POOL.store_instance_ids:
        raise ValueError(f'Unknown Data Store "{store_id}".')
    if store_id is None and _OUTPUT_STORE is None:
        raise ValueError('No default output store set. '
                         'Must specify target store id.')
    store = ECT_DATA_STORE_POOL.get_store(store_instance_id=store_id)
    assert_instance(store, xcube_store.MutableDataStore)
    data_type = _get_data_type(data)
    writer_id = None
    if format_id:
        writer_ids = store.get_data_writer_ids(data_type)
        supported_format_ids = \
            [writer_id.split(':')[1] for writer_id in writer_ids]
        try:
            format_index = supported_format_ids.index(format_id)
        except ValueError:
            raise ValidationError(
                f'Format "{format_id}" is not supported by data store '
                f'"{store_id}". Must be one of the following: '
                f'{", ".join(supported_format_ids)}'
            )
        writer_id = writer_ids[format_index]

    write_work = 10

    with monitor.starting('Writing dataset', write_work):
        with add_progress_observers(
                XcubeProgressObserver(ChildMonitor(monitor, write_work))
        ):
            dataset_id = store.write_data(
                data=data, data_id=data_id, writer_id=writer_id, replace=replace
            )

    return dataset_id


def _get_data_type(data: Any) -> str:
    if isinstance(data, xr.Dataset):
        return xcube_store.DATASET_TYPE.alias
    elif isinstance(data, MultiLevelDataset):
        return xcube_store.MULTI_LEVEL_DATASET_TYPE.alias
    elif isinstance(data, gpd.GeoDataFrame):
        return xcube_store.GEO_DATA_FRAME_TYPE.alias
    else:
        raise ValidationError('Cannot determine data type.')


def get_search_params(store_id: str, data_type: str) -> Dict:
    """
    Returns potential search parameters that can be used to search for datasets
    in a data store.

    :param store_id: The id of the store which shall be searched for data
    :param data_type: An optional data type to specify the type of data to be
        searched

    :return: A dictionary containing search parameters
    """
    store = ECT_DATA_STORE_POOL.get_store(store_id)
    search_params = store.get_search_params_schema(data_type)
    return search_params.to_dict()


def search(store_id: str, data_type: str = None, **search_params) -> List[Dict]:
    """
    Searches in a data store for data that meet the given search criteria.

    :param store_id: The id of the store which shall be searched for data
    :param data_type: An optional data type to specify the type of data to be
        searched
    :param search_params: Store-specific additional search parameters

    :return: A list of dictionaries providing detailed information about the data
        that meet the specified criteria.
    """
    store = ECT_DATA_STORE_POOL.get_store(store_id)
    data_descriptors = list(store.search_data(
        data_type=data_type, **search_params
    ))
    return [data_descriptor.to_dict() for data_descriptor in data_descriptors]



"""
Test Utility operations
"""

from datetime import timezone
from unittest import TestCase

import numpy as np
import pandas as pd
import xarray as xr

from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.ops.utility import merge
from esa_climate_toolbox.util.misc import object_to_qualified_name


class MergeTest(TestCase):
    def test_nominal(self):
        """
        Test nominal execution
        """
        periods = 5
        time = pd.date_range('2000-01-01', periods=periods, tz=timezone.utc)

        ds_1 = xr.Dataset({'A': (['time'], np.random.randn(periods)),
                           'B': (['time'], np.random.randn(periods)),
                           'time': time})
        ds_2 = xr.Dataset({'C': (['time'], np.random.randn(periods)),
                           'D': (['time'], np.random.randn(periods)),
                           'time': time})
        new_ds = merge(ds_1=ds_1, ds_2=ds_2, ds_3=None, ds_4=None)
        self.assertTrue('A' in new_ds)
        self.assertTrue('B' in new_ds)
        self.assertTrue('C' in new_ds)
        self.assertTrue('D' in new_ds)

        new_ds = merge(ds_1=ds_1, ds_2=ds_1, ds_3=ds_1, ds_4=ds_2)
        self.assertTrue('A' in new_ds)
        self.assertTrue('B' in new_ds)
        self.assertTrue('C' in new_ds)
        self.assertTrue('D' in new_ds)

        new_ds = merge(ds_1=ds_1, ds_2=ds_1, ds_3=ds_1, ds_4=ds_1)
        self.assertIs(new_ds, ds_1)

        new_ds = merge(ds_1=ds_2, ds_2=ds_2, ds_3=ds_2, ds_4=ds_2)
        self.assertIs(new_ds, ds_2)

        ds_3 = xr.Dataset({'E': (['time'], np.random.randn(periods)),
                           'time': time})
        new_ds = merge(ds_1=ds_1, ds_2=ds_2, ds_3=ds_3, ds_4=None)
        self.assertTrue('A' in new_ds)
        self.assertTrue('B' in new_ds)
        self.assertTrue('C' in new_ds)
        self.assertTrue('D' in new_ds)
        self.assertTrue('E' in new_ds)

        ds_4 = xr.Dataset({'F': (['time'], np.random.randn(periods)),
                           'time': time})
        new_ds = merge(ds_1=ds_1, ds_2=ds_2, ds_3=ds_3, ds_4=ds_4)
        self.assertTrue('A' in new_ds)
        self.assertTrue('B' in new_ds)
        self.assertTrue('C' in new_ds)
        self.assertTrue('D' in new_ds)
        self.assertTrue('E' in new_ds)

    def test_failures(self):
        with self.assertRaises(ValidationError):
            merge(ds_1=None, ds_2=None, ds_3=None, ds_4=None)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Implementation of custom complex types and appropriate value validation and
conversion routines.

For example::

@op
@op_input('file', data_type=PathLike)
def some_op(file: PathLike.TYPE) -> bool:
    # Do something useful
    pass

"""

import ast
import io
from abc import ABCMeta, abstractmethod
from datetime import datetime, date
from typing import Generic, TypeVar, Union, Optional, Any, Tuple, List

import geopandas
import pandas
import shapely
import shapely.geometry
import shapely.geometry.base
import shapely.wkt
import xarray
from shapely.errors import ShapelyError

from ..util.misc import to_list, to_datetime_range, to_datetime
from ..util.safe import safe_eval

__author__ = "Janis Gailis (S[&]T Norway), " \
             "Norman Fomferra (Brockmann Consult GmbH), " \
             "Marco Zühlke (Brockmann Consult GmbH)"

T = TypeVar('T')


class ValidationError(ValueError):
    """
    Special ``ValueError`` used to signal failed validation of user-provided
    values.
    """

    # noinspection PyArgumentList
    def __init__(self, *args, hint=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.hint = hint


class Like(Generic[T], metaclass=ABCMeta):
    """
    Base class for complex types which can convert a value of varying source
    types into a target type *T*.
    The varying source types are therefore *like* the target type *T*.

    Subclasses shall adhere to the rule that :py:meth:`convert` shall always be
    able to convert from the ``str`` values returned from :py:meth:`format`.
    """

    #: A type that represents the varying source types. This is usually a
    #: ``typing.Union`` instance which combines the varying source types.
    #:  The ``str`` type shall always be among them so that textual value
    #: representations are supported.
    TYPE = Any

    @classmethod
    def name(cls) -> str:
        """Return the name of the type."""
        return cls.__name__

    @classmethod
    def accepts(cls, value: Any) -> bool:
        """
        Return ``True`` if the given value can be converted into the
        target type *T*, ``False`` otherwise.
        """
        try:
            cls.convert(value)
            return True
        except ValueError:
            return False

    @classmethod
    @abstractmethod
    def convert(cls, value: Any) -> Optional[T]:
        """
        Convert the given source value (of type ``Like.TYPE``) into an
        optional instance of type *T*.

        The general contract prescribes that values of type ``str`` shall
        always be allowed. In particular, the ``str`` values returned by the
        :py:meth:`format` method should always be a valid *value*.

        @:raises ValueError if the conversion fails.
        """
        pass

    @classmethod
    def format(cls, value: Optional[T]) -> str:
        """
        Convert the given optional source value of type *T* into a string.

        The general contract prescribes that the value returned shall be
        a valid input to :py:meth:`convert`.

        @:raises ValueError if the conversion fails.
        """
        return str(value)

    @classmethod
    def from_json(cls, value: Any) -> Optional[T]:
        """
        Deserialize the given JSON value into a value of target type *T*.

        :param value: a JSON value
        :return: a optional value of target type *T*
        """
        return cls.convert(value)

    @classmethod
    def to_json(cls, value: Optional[T]) -> Any:
        """
        Serialize the given value of type *T* into a JSON value.

        :param value: an optional value of target type *T*
        :return: a JSON value
        """
        return cls.format(value)

    @classmethod
    def raise_validation_error(cls, value: Any):
        if value is None:
            raise ValidationError("Value of type '{type}' expected."
                                  .format(type=cls.name()))

        text_value = repr(value)
        if len(text_value) > 23:
            text_value = text_value[0:24] + '...'
        raise ValidationError(
            "Value {value} cannot be converted into a '{type}'.".format(
                value=text_value, type=cls.name()
            )
        )


VarNames = List[str]


class Arbitrary(Like[Any]):
    """
    Represents an arbitrary Python value.
    """
    TYPE = Any

    @classmethod
    def convert(cls, value: Any) -> Any:
        """
        Return **value**.
        """
        return value

    @classmethod
    def format(cls, value: Any) -> str:
        if value is None:
            return ''
        return str(value)


class Literal(Like[Any]):
    """
    Represents an arbitrary Python literal.
    """
    TYPE = str

    @classmethod
    def convert(cls, value: Any) -> Any:
        """
        If **value** is a string treat it as a Python literal and return its
        evaluation result, otherwise return **value**.
        """
        if value == '':
            return None
        if isinstance(value, str):
            try:
                return ast.literal_eval(value)
            except (SyntaxError, ValueError):
                cls.raise_validation_error(value)
        return value

    @classmethod
    def format(cls, value: Any) -> str:
        if value is None:
            return ''
        return repr(value)


class HTML(str):
    pass


class HTMLLike(Like[HTML]):
    """
    Represents HTML string
    """
    TYPE = HTML

    @classmethod
    def convert(cls, value: Union[str, HTML]) -> Optional[HTML]:
        """
        Return **value**
        """
        if value is None:
            return None
        return HTML(value)

    @classmethod
    def format(cls, value: HTML) -> Optional[str]:
        if value is None:
            return None
        return value


class VarNamesLike(Like[VarNames]):
    """
    Type class for Variable selection objects

    Accepts:
        1. a string 'pattern1, pattern2, pattern3'
        2. a list ['pattern1', 'pattern2', 'pattern3']

    Converts to a list of strings
    """
    TYPE = Union[VarNames, str]

    @classmethod
    def convert(cls, value: Any) -> Optional[VarNames]:
        """
        Convert the given value to a list of variable name patterns.
        """
        # Can be optional
        if value is None:
            return None

        if isinstance(value, str):
            return to_list(value)

        if not isinstance(value, list):
            raise ValidationError('List of variables names expected.')

        for item in value:
            if not isinstance(item, str):
                raise ValidationError('List of variables names expected.')

        return value.copy()

    @classmethod
    def format(cls, value: Optional[VarNames]) -> str:
        if not value:
            return ''
        if isinstance(value, str):
            return value
        if len(value) == 1:
            return value[0]
        return ', '.join(value)


class VarName(Like[str]):
    """
    Type class for a single Variable selection object.

    Accepts:
        1. a string

    Converts to a string
    """
    TYPE = str

    @classmethod
    def convert(cls, value: Any) -> Optional[str]:
        """
        Convert the given value to a variable name
        """
        # Can be optional
        if value is None:
            return None

        if not isinstance(value, str):
            raise ValidationError('Variable name expected.')

        return value


class DimName(VarName):
    """
    Type class for a single Dimension selection object

    Accepts:
        1. a string

    Converts to a string
    """
    @classmethod
    def convert(cls, value: Any) -> Optional[str]:
        """
        Convert the given value to a variable name
        """
        # Can be optional
        if value is None:
            return None

        if not isinstance(value, str):
            raise ValidationError('Dimension name expected.')

        return value


class DimNamesLike(VarNamesLike):
    """
    Type class for Variable selection objects

    Accepts:
        1. a string 'pattern1, pattern2, pattern3'
        2. a list ['pattern1', 'pattern2', 'pattern3']

    Converts to a list of strings
    """
    @classmethod
    def convert(cls, value: Any) -> Optional[VarNames]:
        """
        Convert the given value to a list of variable name patterns.
        """
        # Can be optional
        if value is None:
            return None

        if isinstance(value, str):
            return to_list(value)

        if not isinstance(value, list):
            raise ValidationError('List of dimension names expected.')

        for item in value:
            if not isinstance(item, str):
                raise ValidationError('List of dimension names expected.')

        return value.copy()


class FileLike(Like[dict]):
    """
    Type class for file-like objects

    Accepts:
        1. a string
        2. an io.IOBase object

    Does not convert at all.
    """

    TYPE = Union[str, io.IOBase]

    @classmethod
    def convert(cls, value: Any) -> Optional[Union[str, io.IOBase]]:
        if not value:
            return None
        if not isinstance(value, str) and not isinstance(value, io.IOBase):
            raise ValidationError('File must be a path or a file handler.')
        return value

    @classmethod
    def format(cls, value: Optional[Union[str, io.IOBase]]) -> str:
        return "{}".format(value) if isinstance(value, str) else ''


class DictLike(Like[dict]):
    """
    Type class for dictionary objects

    Accepts:
        1. a dictionary string
        2. a dict object

    Converts to a dict object
    """

    TYPE = Union[str, dict]

    @classmethod
    def convert(cls, value: Any) -> Optional[dict]:
        if value is None:
            return None

        if isinstance(value, dict):
            return value

        if not isinstance(value, str):
            cls.raise_validation_error(value)

        if value.strip() == '':
            return None

        # noinspection PyBroadException
        try:
            return safe_eval('dict(%s)' % value, {})
        except Exception:
            cls.raise_validation_error(value)

    @classmethod
    def format(cls, value: Optional[dict]) -> str:
        return ', '.join(['%s=%s' % (k, repr(v)) for k, v in value.items()]) \
            if value else ''


class BaseGeometryLike(Generic[T], Like[T], metaclass=ABCMeta):

    @classmethod
    def format(cls, value: shapely.geometry.base.BaseGeometry) -> str:
        return value.wkt if value else ''

    @classmethod
    def to_geometry(cls, value: Any,
                    geom_type: type = shapely.geometry.base.BaseGeometry) \
            -> Optional[shapely.geometry.base.BaseGeometry]:
        if value is None:
            return None

        if isinstance(value, geom_type):
            # Compatible!
            # noinspection PyTypeChecker
            return value

        if isinstance(value, shapely.geometry.base.BaseGeometry):
            # Incompatible!
            raise ValidationError(
                cls.errmsg("Passed geometry type is incompatible.")
            )

        # Try converting to compatible type
        if isinstance(value, str):
            value = value.strip()
            if value == '':
                # Empty strings are same as None
                return None

            try:
                try:
                    coordinates = \
                        [float(v) for v in value.split(',', maxsplit=3)]
                    if len(coordinates) == 2 and cls._is_compatible_type(
                            geom_type, shapely.geometry.Point
                    ):
                        # Point "x, y"?
                        x, y = coordinates
                        return shapely.geometry.Point(x, y)
                    elif len(coordinates) == 4 and cls._is_compatible_type(
                            geom_type, shapely.geometry.Polygon
                    ):
                        # Polygon box "x1, y1, x2, y2"?
                        x1, y1, x2, y2 = coordinates
                        return shapely.geometry.box(x1, y1, x2, y2)
                except (ValueError, TypeError):
                    # Geometry WKT?
                    return shapely.wkt.loads(value)
            except (ShapelyError, ValueError, TypeError) as e:
                raise ValidationError(
                    cls.errmsg('Invalid geometry WKT format.')
                ) from e

            raise ValidationError(cls.errmsg("Unrecognized text format."))

        if cls._is_compatible_type(geom_type, shapely.geometry.Point):
            try:
                # Coordinates list that forms a polygon?
                return shapely.geometry.Point(value)
            except (ShapelyError, ValueError, TypeError):
                pass

        if cls._is_compatible_type(geom_type, shapely.geometry.Polygon):
            try:
                # Coordinates list that forms a polygon?
                return shapely.geometry.Polygon(value)
            except (ShapelyError, ValueError, TypeError):
                # Polygon box x1, y1, x2, y2?
                try:
                    x1, y1, x2, y2 = value
                    return shapely.geometry.box(x1, y1, x2, y2)
                except (ShapelyError, ValueError, TypeError) as e:
                    raise ValidationError(cls.errmsg(str(e))) from e

        raise ValidationError(cls.errmsg())

    @classmethod
    def errmsg(cls, detail=None) -> str:
        if detail:
            return "Value cannot be converted into a '{type}': {detail}".format(
                type=cls.name(), detail=detail
            )
        else:
            return "Value cannot be converted into a '{type}'.".format(
                type=cls.name()
            )

    @classmethod
    def _is_compatible_type(cls, geom_type, required_type):
        return geom_type == shapely.geometry.base.BaseGeometry or issubclass(
            geom_type, required_type
        )


class GeometryLike(BaseGeometryLike[shapely.geometry.base.BaseGeometry]):
    """
    Type class for arbitrary geometry objects

    Accepts:
        1. any Shapely geometry
            (of type ``shapely.geometry.base.BaseGeometry``);
        2. a string 'lon, lat' (a point),
            or 'min_lon, min_lat, max_lon, max_lat' (a shapely.geometry.box);
        3. a Geometry WKT string starting with 'POINT', 'POLYGON', etc;
        4. a coordinate tuple (lon, lat),
            a list of coordinates [(lon, lat), (lon, lat), ...],
            or a list of lists of coordinates [[(lon, lat), (lon, lat), ...],
            [(lon, lat), (lon, lat), ...], ...].

    Converts to a valid shapely geometry.
    """
    TYPE = Union[shapely.geometry.base.BaseGeometry,
                 str,
                 Tuple[float, float],
                 List[Tuple[float, float]],
                 List[List[Tuple[float, float]]]]

    @classmethod
    def convert(cls,
                value: Any) -> Optional[shapely.geometry.base.BaseGeometry]:
        geometry = cls.to_geometry(value, shapely.geometry.base.BaseGeometry)

        if isinstance(geometry, shapely.geometry.Polygon) and geometry.is_valid:
            # Heal polygon, see #506 and Shapely User Manual
            geometry = geometry.buffer(0)

        return geometry


class PointLike(BaseGeometryLike[shapely.geometry.Point]):
    """
    Type class for geometric shapely.geometry.Point objects

    Accepts:
        1. a Shapely shapely.geometry.Point
        2. a string 'lon,lat'
        4. a WKT string "POINT (lon lat)"
        5. a tuple (lon, lat)

    Converts to a Shapely shapely.geometry.Point object.
    """
    TYPE = Union[shapely.geometry.Point, str, Tuple[float, float]]

    @classmethod
    def convert(cls, value: Any) -> Optional[shapely.geometry.Point]:
        return cls.to_geometry(value, shapely.geometry.Point)

    @classmethod
    def format(cls, value: Optional[shapely.geometry.Point]) -> str:
        return "%s, %s" % (value.x, value.y) if value else ''


class PolygonLike(GeometryLike, Like[shapely.geometry.Polygon]):
    """
    Type class for geometric shapely.geometry.Polygon objects

    Accepts:
        1. a ``shapely.geometry.shapely.geometry.Polygon`` object
        2. a string "min_lon, min_lat, max_lon, max_lat"
        3. a WKT string "POLYGON ((RING))" or
            "POLYGON ((OUTER-RING), (INNER-RING), ...)"
        4. a list of coordinates [(lon, lat), (lon, lat), (lon, lat)]
        5. a list or tuple [min_lon, min_lat, max_lon, max_lat]

    Converts to a valid Shapely shapely.geometry.Polygon.
    """
    TYPE = Union[shapely.geometry.Polygon, List[Tuple[float, float]],
                 str, Tuple[float, float, float, float]]

    @classmethod
    def convert(cls, value: Any) -> Optional[shapely.geometry.Polygon]:
        polygon = cls.to_geometry(value, shapely.geometry.Polygon)
        if polygon is None:
            return None

        if not isinstance(polygon, shapely.geometry.Polygon):
            raise ValidationError('Polygon expected.')

        if not polygon.is_valid:
            # Heal polygon, see #506 and Shapely User Manual
            polygon = polygon.buffer(0)
            if not polygon.is_valid:
                raise ValidationError(cls.errmsg("Polygon is invalid."))

        if polygon.is_empty:
            raise ValidationError(cls.errmsg("Polygon is empty."))

        return polygon

    @classmethod
    def from_json(cls, value: Any) -> Optional[T]:
        """
        Deserialize the given JSON value into a value of target type *T*.

        :param value: a JSON value
        :return: a optional value of target type *T*
        """
        if cls.convert(value):
            # We want to preserve tuples if that's how a polygon was defined
            return value
        else:
            return None

    @classmethod
    def format(cls, value: Optional[shapely.geometry.Polygon]) -> str:
        return value.wkt if value else ''


class TimeLike(Like[datetime]):
    """
    Type class for a time-like object.

    Accepts:
        2. a string with format 'YYYY-MM-DD';
        3. a datetime object;
        4. a date object.

    Converts to datetime object.
    """
    TYPE = Union[str, datetime, date]

    @classmethod
    def convert(cls, value: Any) -> Optional[datetime]:
        # Can be optional
        if value is None or isinstance(value, datetime):
            return value

        if isinstance(value, str) and value.strip() == '':
            return None

        if isinstance(value, date) or isinstance(value, str):
            # noinspection PyBroadException
            try:
                return to_datetime(value)
            except Exception:
                pass

        cls.raise_validation_error(value)

    @classmethod
    def format(cls, value: Optional[datetime]) -> str:
        return _to_isoformat(value)


_ZERO_ISO_TIME_PART = 'T00:00:00'


def _to_isoformat(value: datetime) -> str:
    if not value:
        return ''
    text = value.isoformat()
    if text.endswith(_ZERO_ISO_TIME_PART):
        return text[0:-len(_ZERO_ISO_TIME_PART)]
    return text


TimeRange = Tuple[datetime, datetime]


class TimeRangeLike(Like[TimeRange]):
    """
    Type class for temporal selection objects

    Accepts:
        1. a tuple of start/end time datetime strings
            ('YYYY-MM-DD', 'YYYY-MM-DD');
        2. a string of datetime string start/end dates 'YYYY-MM-DD,YYYY-MM-DD';
        3. a tuple of start/end datetime datetime objects;
        4. a tuple of start/end datetime date objects.

    Converts to a tuple of datetime objects.
    """
    TYPE = Union[Tuple[str, str], TimeRange, Tuple[date, date], str]

    @classmethod
    def convert(cls, value: Any) -> Optional[TimeRange]:
        # Can be optional
        if value is None or value == '':
            return None

        if isinstance(value, str):
            value = [x.strip() for x in value.split(',')]

        try:
            t1, t2 = value[0], value[1]
        except IndexError:
            cls.raise_validation_error(value)
            return

        time_range = to_datetime_range(t1, t2)

        # Check if start date is before end date
        if not time_range or time_range[0] <= time_range[1]:
            return time_range

        cls.raise_validation_error(value)

    @classmethod
    def format(cls, value: Optional[TimeRange]) -> str:
        if not value:
            return ''
        return '{}, {}'.format(_to_isoformat(value[0]), _to_isoformat(value[1]))


class DatasetLike(Like[xarray.Dataset]):
    """
    Accepts xarray.Dataset, pandas.DataFrame and converts to xarray.Dataset.
    """

    TYPE = Optional[Union[xarray.Dataset, pandas.DataFrame]]

    @classmethod
    def convert(cls, value: Any) -> Optional[xarray.Dataset]:
        # Can be optional
        from esa_climate_toolbox.core.opimpl import adjust_temporal_attrs_impl

        if value is None:
            return None
        if isinstance(value, xarray.Dataset):
            return value
        if isinstance(value, pandas.DataFrame):
            return adjust_temporal_attrs_impl(
                xarray.Dataset.from_dataframe(value)
            )
        raise ValidationError(
            'Value must be an xarray.Dataset or pandas.DataFrame.'
        )

    @classmethod
    def format(cls, value: Optional[xarray.Dataset]) -> str:
        if value is None:
            return ''
        raise ValidationError(
            'Values of type DatasetLike cannot be converted to text.'
        )


class DataFrameLike(Like[pandas.DataFrame]):
    """
    Accepts pandas.DataFrame, xarray.Dataset and converts to pandas.DataFrame.
    """

    TYPE = Optional[Union[pandas.DataFrame, xarray.Dataset]]

    @classmethod
    def convert(cls, value: Any) -> Optional[pandas.DataFrame]:
        # Can be optional
        if value is None:
            return None
        if isinstance(value, pandas.DataFrame):
            return value
        if isinstance(value, xarray.Dataset):
            return value.to_dataframe()
        raise ValidationError(
            'Value must be a pandas.DataFrame or xarray.Dataset.'
        )

    @classmethod
    def format(cls, value: Optional[pandas.DataFrame]) -> str:
        if value is None:
            return ''
        raise ValidationError(
            'Values of type DataFrameLike cannot be converted to text.'
        )


class GeoDataFrame:
    """
    Proxy for a ``geopandas.GeoDataFrame`` that holds an iterable of features
    or a feature collection for fastest possible streaming of GeoJSON features.

    Important note: although GeoDataFrame is not inherited from
    geopandas.GeoDataFrame it is an instance of that type meaning that::

        isinstance(GeoDataFrame(...), geopandas.GeoDataFrame) == True

    It also has all attributes of a geopandas.GeoDataFrame.
    This allows us using the proxy for operation parameters
    of type parameters pandas.DataFrame and geopandas.GeoDataFrame.
    """

    _OWN_PROPERTY_SET = {
        "_features",
        "features",
        "_lazy_data_frame",
        "lazy_data_frame",
        "close",
    }

    @classmethod
    def from_features(cls, features):
        """
        Create GeoDataFrame from an iterable of features
        or a feature collection.

        :param features: iterable of features or a feature collection,
            e.g., an open ``fiona.Collection`` instance.
        :return: An instance of a ``GeoDataFrame`` proxy.
        """
        return cls(features)

    def __init__(self, features):
        if features is None:
            raise ValueError('features must not be None')
        self._features = features
        self._lazy_data_frame = None

    @property
    def features(self):
        return self._features

    @property
    def lazy_data_frame(self):
        features = self._features
        if features is not None and self._lazy_data_frame is None:
            crs = features.crs if hasattr(features, 'crs') else None
            self._lazy_data_frame = geopandas.GeoDataFrame.from_features(
                features, crs=crs
            )

        return self._lazy_data_frame

    def close(self):
        """
        In the ESA Climate Toolbox, closable resources are closed when removed
        from the resources cache. Therefore we provide a close method here,
        although geopandas.GeoDataFrame doesn't have one.
        """
        try:
            self._features.close()
        except AttributeError:
            pass
        self._features = None
        self._lazy_data_frame = None

    def __setattr__(self, key, value):
        if key in GeoDataFrame._OWN_PROPERTY_SET:
            object.__setattr__(self, key, value)
        else:
            self.lazy_data_frame.__setattr__(key, value)

    def __getattribute__(self, item):
        if item in GeoDataFrame._OWN_PROPERTY_SET:
            return object.__getattribute__(self, item)
        else:
            return getattr(self.lazy_data_frame, item)

    def __getattr__(self, item):
        if item in GeoDataFrame._OWN_PROPERTY_SET:
            return object.__getattribute__(self, item)
        else:
            return getattr(self.lazy_data_frame, item)

    def __getitem__(self, item):
        return self.lazy_data_frame.__getitem__(item)

    def __setitem__(self, key, value):
        return self.lazy_data_frame.__setitem__(key, value)

    def __str__(self):
        return str(self.lazy_data_frame)

    def __repr__(self):
        return repr(self.lazy_data_frame)

    def __len__(self):
        return len(self._features)


GeoDataFrameProxy = GeoDataFrame



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import concurrent.futures
import platform
import re
import subprocess
import shlex
import time
from typing import Callable, Optional, Tuple, Union, Dict, Sequence

from .monitor import Monitor

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"


def run_subprocess(
        command: Union[str, Sequence[str]],
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        shell: bool = False,
        started_handler: Optional[Callable[[subprocess.Popen], None]] = None,
        stdout_handler: Optional[Callable[[str], None]] = None,
        stderr_handler: Optional[Callable[[str], None]] = None,
        done_handler: Optional[Callable[[int], None]] = None,
        is_cancelled: Optional[Callable[[], bool]] = None,
        cancelled_check_period: float = 0.1,
        kill_on_cancel=False
):
    """
    Execute a child program in a new process and wait for its termination.

    :param command: The command to be executed, may be a string or sequence
        of string arguments.
    :param cwd: Optional current working directory.
    :param env: Optional dictionary of environment variables.
    :param shell: Whether to use the shell as the program to execute.
    :param started_handler: An optional callable that is called with the
        program's process (a ``subprocess.Popen`` object)
    :param stdout_handler: An optional callable that receives the process'
        stdout as lines of UTF-8 encoded text.
    :param stderr_handler: An optional callable that receives the process'
        stderr as lines of UTF-8 encoded text.
    :param done_handler: An optional callable that is called with the program's
        exit code
    :param is_cancelled: An optional callable that is called to determine
        whether the program's process should be killed
    :param cancelled_check_period: The time to sleep between subsequent
        *is_cancelled()* calls. Defaults to 0.1 seconds.
    :param kill_on_cancel: Whether to send a SIGKILL rather than a SIGTERM
        signal when cancellation is requested (Unix only)
    :return: the program's return code (an `int`) or `None`
        if it could not be determined.
    """

    assert command, "command must be provided"

    if isinstance(command, str) and not shell:
        args = shlex.split(command, posix=platform.system() != 'Windows')
    else:
        args = command

    process = subprocess.Popen(args,
                               shell=shell,
                               cwd=cwd,
                               env=env,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               bufsize=0)

    if started_handler:
        started_handler(process)

    def _read_line(fp, handler):
        while process.returncode is None:
            line = fp.readline()
            if handler:
                # noinspection PyBroadException
                try:
                    handler(line.decode("utf-8"))
                except Exception:
                    pass
            if not line:
                return
            if is_cancelled is not None and is_cancelled():
                _cancel(process, kill_on_cancel)

    def _read_line_stdout():
        _read_line(process.stdout, stdout_handler)

    def _read_line_stderr():
        _read_line(process.stderr, stderr_handler)

    def _check_cancelled():
        if is_cancelled is None:
            return
        while process.returncode is None:
            if is_cancelled():
                _cancel(process, kill_on_cancel)
            time.sleep(cancelled_check_period or 0.1)

    def _wait():
        return_code = process.wait()
        if done_handler:
            done_handler(return_code)
        return return_code

    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        f1 = executor.submit(_read_line_stdout)
        f2 = executor.submit(_read_line_stderr)
        f3 = executor.submit(_check_cancelled)
        f4 = executor.submit(_wait)
        statuses = concurrent.futures.wait([f1, f2, f3, f4])
        result = f4.result() if f4 in statuses.done else None

    return result


def _cancel(process: subprocess.Popen, kill_on_cancel: bool):
    if kill_on_cancel:
        process.kill()
    else:
        process.terminate()
    return True


class ProcessOutputMonitor:
    r"""
    A stdout handler for :py:func:`execute` the delegates extracted
    progress information to a monitor. Information is extracted using regular
    expressions or a callable that extracts the information.

    :param monitor: The progress monitor to which extracted information from
        process outputs is delegated.
    :param label: A default label that is used in case no start label
        has been extracted
    :param total_work: Default total work that is used in case no total work
        could be determined
    :param started: Either a callable that receives a text line and
        returns a tuple (label, total_work) or a regex that must match in order
        to signal the start of progress monitoring. The regex must provide the
        group names "label" or "total_work" or both,
        e.g. "(?P<label>\w+)" or "(?P<total_work>\d+)"
    :param progress: Either a callable that receives a text line and returns a
        tuple (work, msg) or a regex that must match in order to signal process.
        The regex must provide group names "work" or "msg" or both,
        e.g. "(?P<msg>\w+)" or "(?P<work>\d+)"
    :param done: Either a callable that receives a text line and returns
        True or False or a regex that must match in order to signal the end
        of progress monitoring.
    """

    def __init__(self,
                 monitor: Monitor,
                 label: str = None,
                 total_work: float = None,
                 started: Union[str, Callable] = None,
                 progress: Union[str, Callable] = None,
                 done: Union[str, Callable] = None,
                 work_is_incremental=False):
        self.monitor = monitor
        self.label = label
        self.total_work = total_work
        self.current_work = 0
        self.work_is_incremental = work_is_incremental
        self.has_started = False
        self.is_done = False
        self.re_started = re.compile(started) \
            if isinstance(started, str) \
            else None
        self.re_progress = re.compile(progress) \
            if isinstance(progress, str) \
            else None
        self.re_done = re.compile(done) if isinstance(done, str) else None
        self.parse_started = started if callable(started) else None
        self.parse_progress = progress if callable(progress) else None
        self.parse_done = done if callable(done) else None

    def __call__(self, line: str):
        if self.is_done:
            return

        monitor = self.monitor

        is_done = self._do_parse_done(line)
        if is_done:
            if self.has_started:
                monitor.done()
            self.done = True
            return

        work, msg = self._do_parse_progress(line)
        if work or msg:
            if not self.has_started:
                monitor.start(
                    self.label or 'Process running', total_work=self.total_work
                )
                self.has_started = True
            if work and not self.work_is_incremental:
                work_delta = work - self.current_work
                self.current_work += work_delta
                work = work_delta
            monitor.progress(work=work, msg=msg)

        if not self.has_started:
            label, total_work = self._do_parse_started(line)
            if label or total_work is not None:
                monitor.start(label or self.label or 'Process running',
                              total_work=total_work or self.total_work)
                self.has_started = True

    def _do_parse_started(
            self, line: str
    ) -> Tuple[Optional[str], Optional[float]]:
        label = None
        total_work = None
        if self.re_started:
            m = self.re_started.match(line)
            if m:
                d = m.groupdict()
                label = d.get('label')
                total_work = d.get('total_work')
                if total_work:
                    total_work = float(total_work)
        elif self.parse_started:
            label, total_work = self.parse_started(line)
        return label, total_work

    def _do_parse_progress(
            self, line: str
    ) -> Tuple[Optional[float], Optional[str]]:
        work = None
        msg = None
        if self.re_progress:
            m = self.re_progress.match(line)
            if m:
                d = m.groupdict()
                work = d.get('work')
                if work:
                    work = float(work)
                msg = d.get('msg')
        elif self.parse_progress:
            work, msg = self.parse_progress(line)
        return work, msg

    def _do_parse_done(self, line: str) -> bool:
        if self.re_done:
            return self.re_done.match(line) is not None
        elif self.parse_done:
            return self.parse_done(line)
        else:
            return line == ''



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

import inspect
import re
from collections import OrderedDict
from typing import Tuple, Dict, List, Any, Optional

from .misc import object_to_qualified_name, qualified_name_to_object
from ..core.types import Like

Props = Dict[str, Any]

_SPHINX_PARAM_DIRECTIVE_PATTERN = re.compile(
    ":param (?P<name>[^:]+): (?P<desc>[^:]+)"
)
_SPHINX_RETURN_DIRECTIVE_PATTERN = re.compile(":returns?: (?P<desc>[^:]+)")

_TYPING_PREFIX = 'typing.'


class OpMetaInfo:
    """
    Represents meta-information about an operation:

    * :py:attr:`qualified_name`: a an ideally unique, qualified operation name
    * :py:attr:`header`: dictionary of arbitrary operation attributes
    * :py:attr:`input`: ordered dictionary of named inputs,
      each mapping to a dictionary of arbitrary input attributes
    * :py:attr:`output`: ordered dictionary of named outputs,
      each mapping to a dictionary of arbitrary output attributes

    Warning: `OpMetaInfo`` objects should be considered immutable. However, the
    dictionaries mentioned above are returned "as-is", mostly for performance
    reasons. Changing entries in these dictionaries directly may cause unwanted
    side-effects.

    :param qualified_name: The operation's qualified name.
    :param has_monitor: Whether the operation supports a :py:class:`Monitor`
        keyword argument named ``monitor``.
    :param header: Header information dictionary.
    :param input_names: Input information dictionary.
    :param inputs: Input information dictionary.
    :param outputs: Output information dictionary.
    """

    def __init__(self,
                 qualified_name: str,
                 has_monitor: bool = False,
                 header: dict = None,
                 input_names: List[str] = None,
                 inputs: Dict[str, Props] = None,
                 outputs: Dict[str, Props] = None):
        if not qualified_name:
            raise ValueError("qualified_name must be given")
        self._qualified_name = qualified_name
        self._has_monitor = True if has_monitor else False
        self._header = header if header else dict()
        self._inputs = OrderedDict(inputs if inputs else {})
        self._outputs = OrderedDict(outputs if outputs else {})
        self._input_names = input_names or self._get_input_names(self._inputs)

    #: The constant ``'monitor'``, which is the name of an operation input that
    #: will receive a :py:class:`Monitor` object as value.
    MONITOR_INPUT_NAME = 'monitor'

    #: The constant ``'return'``, which is the name of a single, unnamed
    #: operation output.
    RETURN_OUTPUT_NAME = 'return'

    @property
    def qualified_name(self) -> str:
        """
        :return: Fully qualified name of the actual operation.
        """
        return self._qualified_name

    @property
    def header(self) -> Dict[str, Any]:
        """
        :return: Operation header attributes.
        """
        return self._header

    @property
    def input_names(self) -> List[str]:
        """
        The input names in the order they have been declared.

        :return: List of input names.
        """
        return self._input_names

    @property
    def inputs(self) -> Dict[str, Props]:
        """
        Mapping from an input name to a dictionary of properties describing the
        input.

        :return: Named inputs.
        """
        return self._inputs

    @property
    def outputs(self) -> Dict[str, Props]:
        """
        Mapping from an output name to a dictionary of properties describing the
        output.

        :return: Named outputs.
        """
        return self._outputs

    @property
    def has_monitor(self) -> bool:
        """
        :return: ``True`` if the operation supports a :py:class:`Monitor`
            value as additional keyword argument named ``monitor``.
        """
        return self._has_monitor

    @property
    def has_named_outputs(self) -> bool:
        """
        :return: ``True`` if the output value of the operation is expected be a
            dictionary-like mapping of output names to output values.
        """
        return not (len(self._outputs) == 1 and
                    self.RETURN_OUTPUT_NAME in self._outputs)

    @property
    def can_cache(self) -> bool:
        return not self._header.get('no_cache', False)

    def to_json_dict(self, data_type_to_json=None) -> Dict[str, Any]:
        """
        Return a JSON-serializable dictionary representation of this object.
        E.g. values of the `data_type`` property are converted from Python types
        to their string representation.

        :return: A JSON-serializable dictionary
        """
        json_dict = OrderedDict()
        json_dict['qualified_name'] = self.qualified_name
        if self.has_monitor:
            json_dict['has_monitor'] = True
        if self.header:
            json_dict['header'] = dict(self.header)
        json_dict['inputs'] = self.object_dict_to_json_dict(
            self.inputs, data_type_to_json
        )
        json_dict['outputs'] = self.object_dict_to_json_dict(
            self.outputs, data_type_to_json
        )
        return json_dict

    @classmethod
    def from_json_dict(
            cls, json_dict: Dict[str, Any], json_to_data_type=None, **kwargs
    ) -> 'OpMetaInfo':
        qualified_name = json_dict.get(
            'qualified_name', kwargs.get('qualified_name', None)
        )
        header_obj = json_dict.get('header', kwargs.get('header', None))
        has_monitor = json_dict.get(
            'has_monitor', kwargs.get('has_monitor', False)
        )
        input_dict = json_dict.get('inputs', kwargs.get('inputs', None))
        output_dict = json_dict.get('outputs', kwargs.get('outputs', None))
        return OpMetaInfo(
            qualified_name,
            header=header_obj,
            has_monitor=has_monitor,
            inputs=cls.json_dict_to_object_dict(input_dict, json_to_data_type),
            outputs=cls.json_dict_to_object_dict(output_dict, json_to_data_type)
        )

    @classmethod
    def object_dict_to_json_dict(cls, obj_dict, data_type_to_json=None):
        if not data_type_to_json:
            data_type_to_json = object_to_qualified_name
        json_dict = OrderedDict()
        for name, properties in obj_dict.items():
            json_dict[name] = dict(properties)
            if 'data_type' in properties:
                json_dict[name]['data_type'] = \
                    data_type_to_json(properties['data_type'])
        return json_dict

    @classmethod
    def json_dict_to_object_dict(cls, json_dict, json_to_data_type=None):
        if not json_to_data_type:
            json_to_data_type = qualified_name_to_object
        obj_dict = OrderedDict()
        for name, properties in json_dict.items():
            obj_dict[name] = dict(properties)
            if 'data_type' in properties:
                obj_dict[name]['data_type'] = \
                    json_to_data_type(properties['data_type'])
        return obj_dict

    def __str__(self):
        return "OpMetaInfo('%s')" % self.qualified_name

    def __repr__(self):
        return "OpMetaInfo('%s')" % self.qualified_name

    @classmethod
    def introspect_operation(cls, operation) -> 'OpMetaInfo':
        if not operation:
            raise ValueError("'operation' argument must be given")

        op_qualified_name = object_to_qualified_name(operation, fail=True)

        arg_inputs, input_dict, has_monitor = [], OrderedDict(), False
        if hasattr(operation, '__code__'):
            arg_inputs, input_dict, has_monitor = \
                cls._introspect_inputs_from_callable(operation, False)

        return_name = OpMetaInfo.RETURN_OUTPUT_NAME

        output_dict = OrderedDict()
        if hasattr(operation, '__annotations__'):
            # mapping of parameters names to annotations;
            # 'return' key is reserved for return annotations.
            annotations = operation.__annotations__
            for annotated_name, annotated_type in annotations.items():
                if annotated_name == 'return':
                    # op_meta_info.output can't be present so far
                    # -> assign new dict
                    output_dict[return_name] = dict(data_type=annotated_type)
                elif annotated_name != cls.MONITOR_INPUT_NAME:
                    # input_dict[annotated_name] should be present through
                    # _introspect_inputs_from_callable() call
                    input_dict[annotated_name]['data_type'] = annotated_type
        if len(output_dict) == 0:
            output_dict[return_name] = dict()

        header = dict()
        # Introspect the operation instance
        # (see https://docs.python.org/3.5/library/inspect.html)
        if hasattr(operation, '__doc__'):
            # documentation string
            docstring = operation.__doc__
            if docstring:
                description, param_descriptions, return_description = \
                    cls._parse_docstring(docstring)
                if description:
                    header['description'] = description
                if param_descriptions:
                    for param_name, param_description in \
                            param_descriptions.items():
                        if param_name in input_dict:
                            input_dict[param_name]['description'] = \
                                param_descriptions[param_name]
                if return_description and return_name in output_dict:
                    output_dict[return_name]['description'] = return_description

        return OpMetaInfo(op_qualified_name,
                          header=header,
                          has_monitor=has_monitor,
                          inputs=input_dict,
                          outputs=output_dict)

    @classmethod
    def _introspect_inputs_from_callable(cls, operation, is_method: bool) -> \
            Tuple[List[Tuple[str, Props]], Props, bool]:
        arg_inputs = []
        input_dict = OrderedDict()
        has_monitor = False
        # code object containing compiled function bytecode
        if not hasattr(operation, '__code__'):
            # Check: throw exception here?
            return arg_inputs, input_dict, has_monitor
        code = operation.__code__
        # number of arguments (not including * or ** args)
        arg_count = code.co_argcount
        # tuple of names of arguments and local variables
        arg_names = code.co_varnames[0:arg_count]
        if len(arg_names) > 0 and is_method and arg_names[0] == 'self':
            arg_names = arg_names[1:]
            arg_count -= 1
        # Reserve input slots for input names, but 'monitor'
        arg_pos = 0
        for arg_name in arg_names:
            if cls.MONITOR_INPUT_NAME != arg_name:
                input_props = dict(position=arg_pos)
                arg_inputs.append((arg_name, input_props))
                input_dict[arg_name] = input_props
                arg_pos += 1
            else:
                has_monitor = True
        # Set 'default_value' for input names, but 'monitor'
        if operation.__defaults__:
            # tuple of any default values for positional or keyword parameters
            default_values = operation.__defaults__
            num_default_values = len(default_values)
            for i in range(num_default_values):
                arg_name = arg_names[i - num_default_values]
                if cls.MONITOR_INPUT_NAME != arg_name:
                    default_value = default_values[i]
                    input_dict[arg_name]['default_value'] = default_value
                    if default_value is None:
                        input_dict[arg_name]['nullable'] = True
                    else:
                        input_dict[arg_name]['data_type'] = type(default_value)
        return arg_inputs, input_dict, has_monitor

    def set_default_input_values(self, input_values: Dict):
        """
        If any missing input value in *input_values*,
        set value of "default_value" property, if it exists.

        :param input_values: The dictionary of input values that will be
            modified.
        """
        for name, properties in self.inputs.items():
            if name not in input_values and 'default_value' in properties:
                input_values[name] = properties['default_value']

    def validate_input_values(self,
                              input_values: Dict,
                              except_types=None,
                              validation_exception_class=ValueError):
        """
        Validate given *input_values* against the operation's input properties.

        :param input_values: The dictionary of input values.
        :param except_types: A set of types or ``None``. If an input value's
            type is in this set, it will not be validated against the various
            input properties, such as ``data_type``, ``nullable``,
            ``value_set``, ``value_range``.
        :param validation_exception_class: The exception class to be used to
            raise exceptions if validation fails. Must derive from
            ``BaseException``. Defaults to ``ValueError``.
        :raise validation_error_class: If *input_values* are invalid w.r.t. to
            the operation's input properties.
        """
        inputs = self.inputs
        # Ensure required input values have values (even None is a value).
        for name, properties in inputs.items():
            has_no_default = 'default_value' not in properties
            is_auto = 'context' in properties
            required = has_no_default and not is_auto
            if required and (name not in input_values):
                raise validation_exception_class(
                    "Input '%s' for operation '%s' must be given." %
                    (name, self.qualified_name)
                )
        # Ensure all input values are valid w.r.t. input properties
        for name, value in input_values.items():
            if name not in inputs:
                raise validation_exception_class(
                    "'%s' is not an input of operation '%s'." %
                    (name, self.qualified_name))
            if except_types and type(value) in except_types:
                continue
            input_properties = inputs[name]
            if input_properties.get('context'):
                # Context values will be set by framework
                continue
            data_type = input_properties.get('data_type')
            if value is None:
                default_is_none = input_properties.get('default_value', 1) \
                                  is None
                value_set = input_properties.get('value_set')
                value_set_has_none = value_set and (None in value_set)
                nullable = input_properties.get('nullable', False)
                if not (default_is_none or value_set_has_none or nullable):
                    raise validation_exception_class(
                        "Input '%s' for operation '%s' must be given." %
                        (name, self.qualified_name))
                continue
            if data_type:
                self._validate_value_against_data_type(
                    data_type, value, self.qualified_name, "Input", name,
                    validation_exception_class
                )
            value_set = input_properties.get('value_set', None)
            if value_set and (value not in value_set):
                raise validation_exception_class(
                    "Input '%s' for operation '%s' must be one of %s." % (
                        name, self.qualified_name, value_set))
            value_range = input_properties.get('value_range', None)
            if value_range and (value is None or not (
                    value_range[0] <= value <= value_range[1])
            ):
                raise validation_exception_class(
                    "Input '%s' for operation '%s' must be in range %s." % (
                        name, self.qualified_name, value_range)
                )

    def validate_output_values(self, output_values: Dict,
                               validation_exception_class: type = ValueError):
        """
        Validate given *output_values* against the operation's output
        properties.

        :param output_values: The dictionary of output values.
        :param validation_exception_class: The exception class to be
            used to raise exceptions if validation fails. Must derive from
            ``BaseException``. Defaults to ``ValueError``.
        :raise validation_error_class: If *output_values* are invalid
            w.r.t. to the operation's output properties.
        """
        outputs = self.outputs
        for name, value in output_values.items():
            if name not in outputs:
                raise validation_exception_class(
                    "'%s' is not an output of operation '%s'." %
                    (name, self.qualified_name)
                )
            output_properties = outputs[name]
            if value is not None:
                data_type = output_properties.get('data_type', None)
                if data_type:
                    self._validate_value_against_data_type(
                        data_type, value, self.qualified_name, "Output", name,
                        validation_exception_class
                    )

    @classmethod
    def _validate_value_against_data_type(cls,
                                          data_type,
                                          value,
                                          op_name: str,
                                          port_type: str,
                                          port_name: str,
                                          validation_exception_class: type):
        try:
            value, can_convert = cls._convert_value(data_type, value)
        except (ValueError, validation_exception_class) as e:
            raise validation_exception_class(
                "%s '%s' for operation '%s': %s" %
                (port_type, port_name, op_name, str(e))
            )
        if not can_convert and value is not None:
            is_float_type = data_type is float \
                            and (isinstance(value, float) or
                                 isinstance(value, int))
            if not is_float_type and not is_instance_of(value, data_type):
                raise validation_exception_class(
                    "%s '%s' for operation '%s' must be of type '%s',"
                    " but got type '%s'." % (
                        port_type, port_name, op_name,
                        data_type.__name__, type(value).__name__)
                )

    @classmethod
    def _convert_value(
            cls, data_type: Any, value: Optional[Any]
    ) -> Tuple[Any, bool]:
        """
        Check if the given type has an "convert(value)" method,
        i.e. our XXXLike types, if so return its result.
        """
        # noinspection PyBroadException
        try:
            return data_type.convert(value), True
        except AttributeError:
            return value, False

    @classmethod
    def _parse_docstring(cls, docstring):
        lines = docstring.split('\n')
        description_lines = []
        directive_lines = []
        in_description = False
        in_directive = False
        param_descriptions = OrderedDict()
        return_descriptions = OrderedDict()
        for line in lines:
            line = line.strip()
            if line.startswith(':'):
                if in_description:
                    in_description = False
                elif in_directive:
                    cls._process_sphinx_directive_lines(directive_lines,
                                                        param_descriptions,
                                                        return_descriptions)
                directive_lines = [line]
                in_directive = True
            elif line:
                if in_description:
                    description_lines.append(line)
                elif in_directive:
                    directive_lines.append(line)
                else:
                    description_lines = [line]
                    in_description = True
            else:
                if in_description:
                    description_lines.append('')
                elif in_directive:
                    cls._process_sphinx_directive_lines(directive_lines,
                                                        param_descriptions,
                                                        return_descriptions)
                    directive_lines = []
                    in_directive = False

        if in_directive:
            cls._process_sphinx_directive_lines(directive_lines,
                                                param_descriptions,
                                                return_descriptions)

        return ('\n'.join(description_lines).strip(' \n\t')
                if description_lines else None,
                param_descriptions,
                return_descriptions.get('return', None))

    @classmethod
    def _process_sphinx_directive_lines(
            cls, directive_lines: List[str], param_descriptions: Dict[str, str],
            return_description: Dict[str, str]
    ) -> None:
        # print("consume", annotation_lines)
        text = ' '.join(directive_lines).strip()
        matcher = _SPHINX_PARAM_DIRECTIVE_PATTERN.match(text)
        if matcher:
            name = matcher.group('name')
            text = cls._strip_sphinx_directive_text(matcher.group('desc'))
            param_descriptions[name] = text
            return

        matcher = _SPHINX_RETURN_DIRECTIVE_PATTERN.match(text)
        if matcher:
            text = cls._strip_sphinx_directive_text(matcher.group('desc'))
            return_description['return'] = text

    @classmethod
    def _strip_sphinx_directive_text(cls, description: str) -> str:
        return description.strip().replace('``', '"') \
            .replace(':py:class:', '') \
            .replace(':py:func:', '') \
            .replace(':py:attr:', '') \
            .replace('`', '')

    @classmethod
    def _get_input_names(cls, inputs: Dict[str, Props]) -> List[str]:
        num_inputs = len(inputs)
        if not num_inputs:
            return []
        input_names = num_inputs * ['']
        index = 0
        for name, props in inputs.items():
            position = props.get('position', index)
            if 0 <= position < num_inputs:
                if input_names[position]:
                    raise ValueError(
                        "illegal input property, "
                        "position={} used twice".format(position)
                    )
            else:
                raise ValueError(
                    "illegal input property, expected position={} to {}, "
                    "but was {}".format(0, num_inputs - 1, position)
                )
            input_names[position] = name
            index += 1
        for position in range(num_inputs):
            if not input_names[position]:
                raise ValueError("illegal input properties, "
                                 "position={} is undefined".format(position))
        return input_names


def is_instance_of(value, data_type) -> bool:
    """
    Tests whether the given *value* is an instance of the given *data_type*.
    The latter may be a type from the Python `typing` module.

    :param value: The value to test
    :param data_type: The data type to test against
    :return: True, if so.
    """

    if inspect.isclass(data_type) and issubclass(data_type, Like):
        data_type = data_type.TYPE

    typing_name = repr(data_type)
    if not typing_name.startswith(_TYPING_PREFIX):
        return isinstance(value, data_type)

    typing_name = typing_name[len(_TYPING_PREFIX):]
    bracket_pos = typing_name.find('[', 1)
    if bracket_pos != -1:
        typing_name = typing_name[0:bracket_pos]

    if typing_name == 'Any':
        return True

    if typing_name == 'Union':
        union_args = data_type.__args__ \
            if hasattr(data_type, '__args__') \
            else None
        if union_args is not None:
            for union_arg in union_args:
                if is_instance_of(value, union_arg):
                    return True
        return False

    if typing_name == 'Callable':
        # Don't go into details of return value and parameters types
        return callable(value)

    typing_origin = data_type.__origin__ \
        if hasattr(data_type, '__origin__') \
        else None
    if typing_origin is not None:
        return is_instance_of(value, typing_origin)
    return False



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module provides general utility operations that wrap specific
``xarray`` functions.

The intention is to make available the ``xarray`` API as a set of general,
domain-independent utility functions.

All operations in this module are tagged with the ``"utility"`` tag.

"""
import xarray as xr

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.types import DatasetLike
from esa_climate_toolbox.core.types import ValidationError


@op(tags=['utility'])
@op_input('ds_1', data_type=DatasetLike)
@op_input('ds_2', data_type=DatasetLike)
@op_input('ds_3', data_type=DatasetLike)
@op_input('ds_4', data_type=DatasetLike)
@op_input('join', value_set=["outer", "inner", "left", "right", "exact"])
@op_input('compat',
          value_set=["identical", "equals", "broadcast_equals", "no_conflicts"]
          )
def merge(ds_1: DatasetLike.TYPE,
          ds_2: DatasetLike.TYPE,
          ds_3: DatasetLike.TYPE = None,
          ds_4: DatasetLike.TYPE = None,
          join: str = 'outer',
          compat: str = 'no_conflicts') -> xr.Dataset:
    """
    Merge up to four datasets to produce a new dataset with combined variables
    from each input dataset.

    This is a wrapper for the ``xarray.merge()`` function.

    For documentation refer to xarray documentation at
    http://xarray.pydata.org/en/stable/generated/xarray.Dataset.merge.html#xarray.Dataset.merge

    The *compat* argument indicates how to compare variables of the same name
    for potential conflicts:

    * "broadcast_equals": all values must be equal when variables are broadcast
      against each other to ensure common dimensions.
    * "equals": all values and dimensions must be the same.
    * "identical": all values, dimensions and attributes must be the same.
    * "no_conflicts": only values which are not null in both datasets must be
        equal. The returned dataset then contains the combination of all
        non-null values.

    :param ds_1: The first input dataset.
    :param ds_2: The second input dataset.
    :param ds_3: An optional 3rd input dataset.
    :param ds_4: An optional 4th input dataset.
    :param join: How to combine objects with different indexes.
    :param compat: How to compare variables of the same name for potential
        conflicts.
    :return: A new dataset with combined variables from each input dataset.
    """

    ds_1 = DatasetLike.convert(ds_1)
    ds_2 = DatasetLike.convert(ds_2)
    ds_3 = DatasetLike.convert(ds_3)
    ds_4 = DatasetLike.convert(ds_4)

    datasets = []
    for ds in (ds_1, ds_2, ds_3, ds_4):
        if ds is not None:
            included = False
            for ds2 in datasets:
                if ds is ds2:
                    included = True
            if not included:
                datasets.append(ds)

    if len(datasets) == 0:
        raise ValidationError('At least two different datasets must be given')
    elif len(datasets) == 1:
        return datasets[0]
    else:
        return xr.merge(datasets, compat=compat, join=join)



# The MIT License (MIT)
# Copyright (c) 2024 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest

from xcube.core.new import new_cube

from esa_climate_toolbox.functions.regions import get_land_mask
from esa_climate_toolbox.functions.regions import get_regions_mask
from esa_climate_toolbox.functions.regions import make_regions_dataset
from esa_climate_toolbox.functions.regions import mask_dataset_by_land
from esa_climate_toolbox.functions.regions import mask_dataset_by_regions


class CountriesTest(unittest.TestCase):

    def setUp(self) -> None:
        self.cube = new_cube(
            width=36, height=18, x_res=10.0, y_res=10.0, variables={'s': 2.0}
        )

    def test_make_regions_dataset(self):
        regions_dataset = make_regions_dataset(self.cube)

        self.assertIsNotNone(regions_dataset)
        self.assertEqual(2, len(regions_dataset.data_vars))
        self.assertIn("country_code", regions_dataset.data_vars)
        self.assertIn("continent_code", regions_dataset.data_vars)

    def test_get_land_mask(self):
        land_mask = get_land_mask(self.cube)

        self.assertIsNotNone(land_mask)
        self.assertEqual(1, len(land_mask.data_vars))
        self.assertIn("land", land_mask.data_vars)

    def test_get_regions_mask(self):
        regions_mask = get_regions_mask(self.cube, regions=["Oceania", "South Africa"])

        self.assertIsNotNone(regions_mask)
        self.assertEqual(1, len(regions_mask.data_vars))
        self.assertIn("regions", regions_mask.data_vars)

    def test_mask_dataset_by_land(self):
        masked_ds = mask_dataset_by_land(self.cube)

        self.assertIsNotNone(masked_ds)
        self.assertEqual(1, len(masked_ds.data_vars))
        self.assertIn("s", masked_ds.data_vars)

    def test_mask_dataset_by_regions(self):
        masked_ds = mask_dataset_by_regions(
            self.cube, regions=["Oceania", "South Africa"]
        )

        self.assertIsNotNone(masked_ds)
        self.assertEqual(1, len(masked_ds.data_vars))
        self.assertIn("s", masked_ds.data_vars)



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

The ``ops`` package comprises all specific operation and processor
implementations.

This is a plugin package automatically imported by the installation script's
entry point ``esa_climate_toolbox_ops`` see the projects ``setup.py`` file).

Verification
============

The module's unit-tests are located in
`test/ops <https://github.com/esa-cci/esa-climate-toolbox/tree/main/test/ops>`_
and may be executed using
``$ py.test test/ops/test_<MODULE>.py --cov=esa-climate-toolbox/ops/<MODULE>.py``
for extra code coverage information.

Functions
==========
"""


# We need ect_init being accessible to use by the plugin registering logic
# before any attempt to import any of the submodules is made.


def ect_init():
    # Plugin initializer.
    # Left empty because operations are registered automatically via decorators.
    pass


from .aggregate import climatology
from .aggregate import reduce
from .aggregate import temporal_aggregation
from .animate import animate_map
from .anomaly import anomaly_external
from .anomaly import anomaly_internal
from .arithmetics import arithmetics
from .arithmetics import diff
from .coregistration import coregister
from .data_frame import add_dataset_values_to_geodataframe
from .data_frame import aggregate_statistics
from .data_frame import as_geodataframe
from .data_frame import data_frame_max
from .data_frame import data_frame_min
from .data_frame import data_frame_subset
from .data_frame import find_closest
from .data_frame import query
from .data_frame import to_dataframe
from .data_frame import to_dataset
from .normalize import adjust_spatial_attrs
from .normalize import adjust_temporal_attrs
from .normalize import normalize
from .outliers import detect_outliers
from .plot import plot
from .plot import plot_contour
from .plot import plot_hist
from .plot import plot_line
from .plot import plot_map
from .plot import plot_scatter
from .resampling import downsample_2d
from .resampling import resample
from .resampling import resample_2d
from .resampling import upsample_2d
from .subset import subset_spatial
from .subset import subset_temporal
from .subset import subset_temporal_index
from .timeseries import tseries_mean
from .timeseries import tseries_point
from.utility import merge

__all__ = [
    # aggregation
    'climatology',
    'reduce',
    'temporal_aggregation',
    # animation
    'animate_map',
    # anomaly
    'anomaly_external',
    'anomaly_internal',
    # arithmetics
    'arithmetics',
    'diff',
    # .coregistration
    'coregister',
    # data frame
    'add_dataset_values_to_geodataframe',
    'aggregate_statistics',
    'as_geodataframe',
    'data_frame_max',
    'data_frame_min',
    'data_frame_subset',
    'find_closest',
    'query',
    'to_dataframe',
    'to_dataset',
    # .normalize
    'normalize',
    'adjust_temporal_attrs',
    'adjust_spatial_attrs',
    # .outliers
    'detect_outliers',
    # .plot
    'plot',
    'plot_contour',
    'plot_hist',
    'plot_line',
    'plot_scatter',
    # .resampling
    'resample',
    'resample_2d',
    'downsample_2d',
    'upsample_2d',
    # .subset
    'subset_spatial',
    'subset_temporal',
    'subset_temporal_index',
    # .timeseries
    'tseries_point',
    'tseries_mean',
    # .utility
    'merge'
]



"""
Test coregistration, checks if the values seem as expected
when using default upsampling/downsampling methods.

"""

from unittest import TestCase

import numpy as np
import xarray as xr
from numpy.testing import assert_almost_equal, assert_array_equal

from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.util.misc import object_to_qualified_name

from esa_climate_toolbox.ops import coregister
from esa_climate_toolbox.ops.coregistration import _find_intersection
from ..util.test_monitor import RecordingMonitor


class TestCoregistration(TestCase):
    """
    Test coregistration
    """

    def test_nominal(self):
        """
        Test nominal execution
        """
        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])
        }).chunk(chunks={'lat': 2, 'lon': 4})

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])
        }).chunk(chunks={'lat': 3, 'lon': 3})

        # Test that the coarse dataset has been resampled onto the grid
        # of the finer dataset.
        rm = RecordingMonitor()
        ds_coarse_resampled = coregister(ds_fine, ds_coarse, monitor=rm)
        self.assertGreaterEqual(len(rm.records), 3)
        self.assertIn('start', rm.records[0])
        self.assertIn('done', rm.records[-1])
        self.assertIn('progress', rm.records[1])

        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array(
                [[[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
                  [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
                  [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
                  [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]],
                 [[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
                  [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
                  [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
                  [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]]])),
            'second': (['time', 'lat', 'lon'], np.array(
                [[[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
                  [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
                  [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
                  [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]],
                 [[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
                  [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
                  [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
                  [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]]])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])
        })
        assert_almost_equal(
            ds_coarse_resampled['first'].values, expected['first'].values
        )

        # Test that the fine dataset has been resampled (aggregated)
        # onto the grid of the coarse dataset.
        ds_fine_resampled = coregister(ds_coarse, ds_fine)
        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array(
                [[[0.625, 0.125, 0., 0., 0., 0.],
                  [0.125, 0.5, 0.125, 0., 0., 0.],
                  [0., 0.125, 0.625, 0., 0., 0.]],
                 [[0.625, 0.125, 0., 0., 0., 0.],
                  [0.125, 0.5, 0.125, 0., 0., 0.],
                  [0., 0.125, 0.625, 0., 0., 0.]]])),
            'second': (['time', 'lat', 'lon'], np.array(
                [[[0.625, 0.125, 0., 0., 0., 0.],
                  [0.125, 0.5, 0.125, 0., 0., 0.],
                  [0., 0.125, 0.625, 0., 0., 0.]],
                 [[0.625, 0.125, 0., 0., 0., 0.],
                  [0.125, 0.5, 0.125, 0., 0., 0.],
                  [0., 0.125, 0.625, 0., 0., 0.]]])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])
        })

        assert_almost_equal(ds_fine_resampled['first'].values, expected['first'].values)

    def test_registered(self):
        """
        Test registered operation execution execution
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(coregister))
        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])
        })

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])
        })

        # Test that the coarse dataset has been resampled onto the grid
        # of the finer dataset.
        ds_coarse_resampled = reg_op(ds_primary=ds_fine, ds_replica=ds_coarse)
        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array(
                [[[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
                  [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
                  [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
                  [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]],
                 [[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
                  [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
                  [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
                  [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]]]
            )),
            'second': (['time', 'lat', 'lon'], np.array(
                [[[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
                  [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
                  [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
                  [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]],
                 [[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
                  [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
                  [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
                  [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]]]
            )),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])})
        assert_almost_equal(
            ds_coarse_resampled['first'].values, expected['first'].values
        )

        # Test that the fine dataset has been resampled (aggregated)
        # onto the grid of the coarse dataset.
        ds_fine_resampled = reg_op(ds_primary=ds_coarse, ds_replica=ds_fine)
        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array(
                [[[0.625, 0.125, 0., 0., 0., 0.],
                  [0.125, 0.5, 0.125, 0., 0., 0.],
                  [0., 0.125, 0.625, 0., 0., 0.]],
                 [[0.625, 0.125, 0., 0., 0., 0.],
                  [0.125, 0.5, 0.125, 0., 0., 0.],
                  [0., 0.125, 0.625, 0., 0., 0.]]]
            )),
            'second': (['time', 'lat', 'lon'], np.array(
                [[[0.625, 0.125, 0., 0., 0., 0.],
                  [0.125, 0.5, 0.125, 0., 0., 0.],
                  [0., 0.125, 0.625, 0., 0., 0.]],
                 [[0.625, 0.125, 0., 0., 0., 0.],
                  [0.125, 0.5, 0.125, 0., 0., 0.],
                  [0., 0.125, 0.625, 0., 0., 0.]]]
            )),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])})

        assert_almost_equal(ds_fine_resampled['first'].values, expected['first'].values)

    def test_error(self):
        """
        Test error conditions
        """
        # Test unexpected global bounds
        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'lat': np.linspace(67.5, 135, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])})

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])})

        with self.assertRaises(ValueError) as err:
            coregister(ds_fine, ds_coarse)
        self.assertIn('(67.5, 135.0)', str(err.exception))

        # Test non-equidistant dataset
        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'lat': [-67.5, -20, 20, 67.5],
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])})

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])})

        with self.assertRaises(ValueError) as err:
            coregister(ds_fine, ds_coarse)
        self.assertIn('not equidistant', str(err.exception))

        # Test non-pixel registered dataset
        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])})

        ds_coarse_err = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.zeros([2, 5, 10])),
            'second': (['time', 'lat', 'lon'], np.zeros([2, 5, 10])),
            'lat': np.linspace(-90, 90, 5),
            'lon': np.linspace(-162, 162, 10),
            'time': np.array([1, 2])})

        with self.assertRaises(ValueError) as err:
            coregister(ds_fine, ds_coarse_err)
        self.assertIn('not pixel-registered', str(err.exception))

        ds_coarse_err = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.zeros([2, 5, 10])),
            'second': (['time', 'lat', 'lon'], np.zeros([2, 5, 10])),
            'lat': np.linspace(-72, 72, 5),
            'lon': np.linspace(-180, 180, 10),
            'time': np.array([1, 2])})

        with self.assertRaises(ValueError) as err:
            coregister(ds_fine, ds_coarse_err)
        self.assertIn('not pixel-registered', str(err.exception))

        # Test unexpected dimensionality
        ds_fine = xr.Dataset({
            'first': (['lat', 'longertude'], np.eye(4, 8)),
            'second': (['lat', 'longertude'], np.eye(4, 8)),
            'lat': np.linspace(-67.5, 67.5, 4),
            'longertude': np.linspace(-157.5, 157.5, 8)})

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])})

        with self.assertRaises(ValueError) as err:
            coregister(ds_fine, ds_coarse)
        self.assertIn('longertude', str(err.exception))

        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])})

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'second': (['time', 'lon'], np.eye(2, 6)),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])})

        with self.assertRaises(ValueError) as err:
            coregister(ds_fine, ds_coarse)
        self.assertIn('select_var', str(err.exception))

    def test_find_intersection(self):
        """
        Test the _find_intersection method
        """

        # Test =======
        #          =========
        a = np.linspace(0.5, 9.5, 10)
        b = np.linspace(5.5, 14.5, 10)
        result = _find_intersection(a, b, (0, 15))
        self.assertEqual((5, 10), result)

        # Test   =======
        #    =========
        a = np.linspace(0.5, 9.5, 10)
        b = np.linspace(5.5, 14.5, 10)
        result = _find_intersection(b, a, (0, 15))
        self.assertEqual((5, 10), result)

        # Test   =======
        #     ==============
        a = np.linspace(5.5, 14.5, 10)
        b = np.linspace(0.5, 19.5, 20)
        result = _find_intersection(a, b, (0, 20))
        self.assertEqual((5, 15), result)

        # Test ==================
        #          ========
        a = np.linspace(0.5, 19.5, 20)
        b = np.linspace(5.5, 14.5, 10)
        result = _find_intersection(a, b, (0, 20))
        self.assertEqual((5, 15), result)

        # Test ============
        #                    ========
        a = np.linspace(0.5, 9.5, 10)
        b = np.linspace(10.5, 19.5, 10)
        with self.assertRaises(ValueError) as err:
            _find_intersection(a, b, (0, 20))
        self.assertIn('valid intersection', str(err.exception))

        # Test       ============
        #  ========
        a = np.linspace(0.5, 9.5, 10)
        b = np.linspace(10.5, 19.5, 10)
        with self.assertRaises(ValueError) as err:
            _find_intersection(b, a, (0, 20))
        self.assertIn('valid intersection', str(err.exception))

        # Test misaligned origins
        a = np.linspace(0.5, 9.5, 10)
        b = np.linspace(1, 9, 10)
        with self.assertRaises(ValueError) as err:
            _find_intersection(a, b, (0, 10))
        self.assertIn('valid intersection', str(err.exception))

        # Test differing pixel sizes
        a = np.linspace(0.5, 9.5, 10)
        b = np.linspace(5.25, 14.75, 20)
        result = _find_intersection(b, a, (0, 20))
        self.assertEqual((5, 10), result)

    def test_subset(self):
        """
        Test coregistration being run on a subset
        """
        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])})

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])})

        lat_slice = slice(-70, 70)
        lon_slice = slice(-40, 40)
        ds_coarse = ds_coarse.sel(lat=lat_slice, lon=lon_slice)

        # Test that the coarse dataset has been resampled onto the grid
        # of the finer dataset.
        ds_coarse_resampled = coregister(ds_fine, ds_coarse)
        assert_array_equal([-67.5, -22.5, 22.5, 67.5], ds_coarse_resampled.lat.values)
        assert_array_equal([-22.5, 22.5],
                           ds_coarse_resampled.lon.values)

        # Check if the geospatial attributes have been correctly set
        self.assertEqual(ds_coarse_resampled.lat.values[0] - 45 * 0.5,
                         ds_coarse_resampled.attrs['geospatial_lat_min'])
        self.assertEqual(ds_coarse_resampled.lat.values[-1] + 45 * 0.5,
                         ds_coarse_resampled.attrs['geospatial_lat_max'])
        self.assertEqual(ds_coarse_resampled.lon.values[0] - 45 * 0.5,
                         ds_coarse_resampled.attrs['geospatial_lon_min'])
        self.assertEqual(ds_coarse_resampled.lon.values[-1] + 45 * 0.5,
                         ds_coarse_resampled.attrs['geospatial_lon_max'])
        self.assertEqual(45.0,
                         ds_coarse_resampled.attrs['geospatial_lat_resolution'])
        self.assertEqual(45.0,
                         ds_coarse_resampled.attrs['geospatial_lon_resolution'])

    def test_recursive(self):
        """
        Test coregistration with more dimensions than lat/lon/time
        """
        slice_fine = np.eye(4, 8)
        slice_coarse = np.eye(3, 6)
        ndarr_fine = np.zeros([2, 2, 2, 4, 8])
        ndarr_coarse = np.zeros([2, 2, 2, 3, 6])
        ndarr_fine_l1 = np.zeros([2, 2, 4, 8])
        ndarr_coarse_l1 = np.zeros([2, 2, 3, 6])
        ndarr_fine_l2 = np.zeros([2, 2, 4, 8])
        ndarr_coarse_l2 = np.zeros([2, 2, 3, 6])
        ndarr_fine[:] = slice_fine
        ndarr_coarse[:] = slice_coarse
        ndarr_fine_l1[:] = slice_fine
        ndarr_coarse_l1[:] = slice_coarse
        ndarr_fine_l2[:] = slice_fine
        ndarr_coarse_l2[:] = slice_coarse

        ds_fine = xr.Dataset({
            'first': (['time', 'layer', 'layer2', 'lat', 'lon'], ndarr_fine),
            'second': (['time', 'layer', 'layer2', 'lat', 'lon'], ndarr_fine),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'layer': np.array([1, 2]),
            'layer2': np.array([1, 2]),
            'time': np.array([1, 2])}).chunk(chunks={'lat': 2, 'lon': 4})

        ds_coarse = xr.Dataset({
            'first': (['time', 'layer', 'layer2', 'lat', 'lon'], ndarr_coarse),
            'second': (['time', 'layer', 'layer2', 'lat', 'lon'], ndarr_coarse),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2]),
            'layer': np.array([1, 2]),
            'layer2': np.array([1, 2])}).chunk(chunks={'lat': 3, 'lon': 3})

        # Test that the coarse dataset has been resampled onto the grid
        # of the finer dataset.
        rm = RecordingMonitor()
        ds_coarse_resampled = coregister(ds_fine, ds_coarse, monitor=rm)
        self.assertGreaterEqual(len(rm.records), 3)
        self.assertIn('start', rm.records[0])
        self.assertIn('done', rm.records[-1])
        self.assertIn('progress', rm.records[1])

        slice_exp = np.array(
            [[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
             [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
             [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
             [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]]
        )
        ndarr_fine_exp = np.zeros([2, 2, 2, 4, 8])
        ndarr_fine_exp[:] = slice_exp

        expected = xr.Dataset({
            'first': (['time', 'layer', 'layer2', 'lat', 'lon'], ndarr_fine_exp),
            'second': (['time', 'layer', 'layer2', 'lat', 'lon'], ndarr_fine_exp),
            'layer': np.array([1, 2]),
            'layer2': np.array([1, 2]),
            'time': np.array([1, 2])})
        assert_almost_equal(
            ds_coarse_resampled['first'].values, expected['first'].values
        )

        # Test that the fine dataset has been resampled (aggregated)
        # onto the grid of the coarse dataset.
        ds_fine_resampled = coregister(ds_coarse, ds_fine)

        slice_exp = np.array([[0.625, 0.125, 0., 0., 0., 0.],
                              [0.125, 0.5, 0.125, 0., 0., 0.],
                              [0., 0.125, 0.625, 0., 0., 0.]])
        ndarr_coarse_exp = np.zeros([2, 2, 2, 3, 6])
        ndarr_coarse_exp[:] = slice_exp

        expected = xr.Dataset({
            'first': (['time', 'layer', 'layer2', 'lat', 'lon'], ndarr_coarse_exp),
            'second': (['time', 'layer', 'layer2', 'lat', 'lon'], ndarr_coarse_exp),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'layer': np.array([1, 2]),
            'layer2': np.array([1, 2]),
            'time': np.array([1, 2])})

        assert_almost_equal(ds_fine_resampled['first'].values, expected['first'].values)

        # Test that coregistering with data arrays with less than all possible
        # dimensions works
        ds_fine = xr.Dataset({
            'first': (['time', 'layer', 'lat', 'lon'], ndarr_fine_l1),
            'second': (['time', 'layer2', 'lat', 'lon'], ndarr_fine_l2),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'layer': np.array([1, 2]),
            'layer2': np.array([1, 2]),
            'time': np.array([1, 2])}).chunk(chunks={'lat': 2, 'lon': 4})

        ds_coarse = xr.Dataset({
            'first': (['time', 'layer', 'lat', 'lon'], ndarr_coarse_l1),
            'second': (['time', 'layer2', 'lat', 'lon'], ndarr_coarse_l2),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2]),
            'layer': np.array([1, 2]),
            'layer2': np.array([1, 2])}).chunk(chunks={'lat': 3, 'lon': 3})

        ds_fine_resampled = coregister(ds_coarse, ds_fine)
        ndarr_coarse_exp = np.zeros([2, 2, 3, 6])
        ndarr_coarse_exp[:] = slice_exp

        expected = xr.Dataset({
            'first': (['time', 'layer', 'lat', 'lon'], ndarr_coarse_exp),
            'second': (['time', 'layer2', 'lat', 'lon'], ndarr_coarse_exp),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'layer': np.array([1, 2]),
            'layer2': np.array([1, 2]),
            'time': np.array([1, 2])})

        assert_almost_equal(ds_fine_resampled['first'].values, expected['first'].values)

    def test_2D(self):
        """
        Test a case where a 2D lat/lon dataset is resampled or used for
        resampling
        """

        # Primary dataset is 2D
        ds_fine = xr.Dataset({
            'first': (['lat', 'lon'], np.eye(4, 8)),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8)}).chunk()

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])}).chunk(chunks={'lat': 3, 'lon': 3})

        ds_coarse_resampled = coregister(ds_fine, ds_coarse)

        slice_exp = np.array(
            [[1., 0.28571429, 0., 0., 0., 0., 0., 0.],
             [0.33333333, 0.57142857, 0.38095238, 0., 0., 0., 0., 0.],
             [0., 0.47619048, 0.52380952, 0.28571429, 0.04761905, 0., 0., 0.],
             [0., 0., 0.42857143, 0.85714286, 0.14285714, 0., 0., 0.]]
        )
        exp_arr = np.zeros([2, 4, 8])
        exp_arr[:] = slice_exp

        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], exp_arr),
            'second': (['time', 'lat', 'lon'], exp_arr),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])})

        assert_almost_equal(
            ds_coarse_resampled['first'].values, expected['first'].values
        )

        # replica dataset contains a 2D variable
        ds_coarse = xr.Dataset({
            'first': (['lat', 'lon'], np.eye(3, 6)),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])}).chunk(chunks={'lat': 3, 'lon': 3})

        ds_coarse_resampled = coregister(ds_fine, ds_coarse)

        assert_almost_equal(ds_coarse_resampled['first'].values, slice_exp)

    def test_int_array(self):
        """
        Test coregistration on integer arrays
        """
        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'],
                      np.array([np.eye(4, 8), np.eye(4, 8)], dtype='int32')),
            'second': (['time', 'lat', 'lon'],
                       np.array([np.eye(4, 8), np.eye(4, 8,)])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])}).chunk(chunks={'lat': 2, 'lon': 4})

        ds_coarse = xr.Dataset({
            'first': (['time', 'lat', 'lon'],
                      np.array([np.eye(3, 6), np.eye(3, 6)], dtype='int32')),
            'second': (['time', 'lat', 'lon'],
                       np.array([np.eye(3, 6), np.eye(3, 6)])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])}).chunk(chunks={'lat': 3, 'lon': 3})

        # Test that the coarse dataset has been resampled onto the grid
        # of the finer dataset.
        ds_coarse_resampled = coregister(ds_fine, ds_coarse, method_us='nearest')

        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([[[1, 1, 0, 0, 0, 0, 0, 0],
                                                         [1, 1, 0, 0, 0, 0, 0, 0],
                                                         [0, 0, 1, 0, 0, 0, 0, 0],
                                                         [0, 0, 0, 1, 0, 0, 0, 0]],
                                                        [[1, 1, 0, 0, 0, 0, 0, 0],
                                                         [1, 1, 0, 0, 0, 0, 0, 0],
                                                         [0, 0, 1, 0, 0, 0, 0, 0],
                                                         [0, 0, 0, 1, 0, 0, 0, 0]]])),
            'second': (['time', 'lat', 'lon'], np.array([[[1, 1, 0, 0, 0, 0, 0, 0],
                                                         [1, 1, 0, 0, 0, 0, 0, 0],
                                                         [0, 0, 1, 0, 0, 0, 0, 0],
                                                         [0, 0, 0, 1, 0, 0, 0, 0]],
                                                         [[1, 1, 0, 0, 0, 0, 0, 0],
                                                         [1, 1, 0, 0, 0, 0, 0, 0],
                                                         [0, 0, 1, 0, 0, 0, 0, 0],
                                                         [0, 0, 0, 1, 0, 0, 0, 0]]])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])})
        assert_almost_equal(
            ds_coarse_resampled['first'].values, expected['first'].values
        )

        # Test that the fine dataset has been resampled (aggregated)
        # onto the grid of the coarse dataset.
        ds_fine_resampled = coregister(ds_coarse, ds_fine, method_ds='mode')
        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([[[1, 0, 0, 0, 0, 0],
                                                         [0, 1, 0, 0, 0, 0],
                                                         [0, 0, 1, 0, 0, 0]],

                                                        [[1, 0, 0, 0, 0, 0],
                                                         [0, 1, 0, 0, 0, 0],
                                                         [0, 0, 1, 0, 0, 0]]])),
            'second': (['time', 'lat', 'lon'], np.array([[[1, 0, 0, 0, 0, 0],
                                                         [0, 1, 0, 0, 0, 0],
                                                         [0, 0, 1, 0, 0, 0]],

                                                        [[1, 0, 0, 0, 0, 0],
                                                         [0, 1, 0, 0, 0, 0],
                                                         [0, 0, 1, 0, 0, 0]]])),
            'lat': np.linspace(-60, 60, 3),
            'lon': np.linspace(-150, 150, 6),
            'time': np.array([1, 2])})

        assert_almost_equal(ds_fine_resampled['first'].values, expected['first'].values)

    def test_same_grid(self):
        """
        Test the case when both datasets already have the same geospatial definition
        """
        ds_fine = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'second': (['time', 'lat', 'lon'], np.array([np.eye(4, 8), np.eye(4, 8)])),
            'lat': np.linspace(-67.5, 67.5, 4),
            'lon': np.linspace(-157.5, 157.5, 8),
            'time': np.array([1, 2])}).chunk(chunks={'lat': 2, 'lon': 4})

        rm = RecordingMonitor()
        ds_same = coregister(ds_fine, ds_fine, monitor=rm)
        # Make sure it returned the input as opposed to going through with
        # coregistration
        self.assertEqual([], rm.records)

        assert_almost_equal(ds_same['first'].values, ds_fine['first'].values)

        # Test that a subset is performed, but no coregistration done
        lat_slice = slice(-70, 70)
        lon_slice = slice(-40, 40)
        ds_subset = ds_fine.sel(lat=lat_slice, lon=lon_slice)

        ds_coreg = coregister(ds_subset, ds_fine, monitor=rm)
        self.assertEqual([], rm.records)
        assert_almost_equal(ds_coreg['first'].values, ds_subset['first'].values)
